from functools import partial

import scipy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection

import plotly.graph_objects as go
from ipywidgets import widgets, fixed, FloatProgress, Layout

from pynwb.epoch import TimeIntervals
from pynwb.misc import AnnotationSeries, Units, DecompositionSeries

from .analysis.spikes import compute_smoothed_firing_rate
from .controllers import (
    make_trial_event_controller,
    GroupAndSortController,
    StartAndDurationController,
    ProgressBar,
)
from .utils.dynamictable import infer_categorical_columns, extract_data_from_intervals
from .utils.mpl import create_big_ax
from .utils.plotly import event_group
from .utils.units import (
    get_spike_times,
    get_max_spike_time,
    get_min_spike_time,
    align_by_time_intervals,
    get_unobserved_intervals,
)
from .utils.widgets import interactive_output, clean_axes

color_wheel = plt.rcParams["axes.prop_cycle"].by_key()["color"]


def show_annotations(annotations: AnnotationSeries, **kwargs):
    fig, ax = plt.subplots()
    ax.eventplot(annotations.timestamps, **kwargs)
    ax.set_xlabel("time (s)")
    return fig


def show_session_raster(
        units: Units,
        time_window=None,
        units_window=None,
        show_obs_intervals=True,
        order=None,
        group_inds=None,
        labels=None,
        show_legend=True,
        progress_bar=None,
):
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    time_window: [int, int]
    units_window: [int, int]
    show_obs_intervals: bool
    order: array-like, optional
    group_inds: array-like, optional
    labels: array-like, optional
    show_legend: bool
        default = True
        Does not show legend if color_by is None or 'id'.
    progress_bar: FloatProgress, optional

    Returns
    -------
    matplotlib.pyplot.Figure

    """

    if time_window is None:
        time_window = [get_min_spike_time(units), get_max_spike_time(units)]

    if units_window is None:
        units_window = [0, len(units)]

    if order is None:
        order = np.arange(len(units), dtype="int")

    if progress_bar:
        this_iter = ProgressBar(order, desc="reading spike data", leave=False)
        progress_bar = this_iter.container
    else:
        this_iter = order
    data = []
    for unit in this_iter:
        data.append(get_spike_times(units, unit, time_window))

    if show_obs_intervals:
        unobserved_intervals_list = get_unobserved_intervals(units, time_window, order)
    else:
        unobserved_intervals_list = None

    ax = plot_grouped_events(
        data,
        time_window,
        group_inds=group_inds,
        labels=labels,
        show_legend=show_legend,
        offset=units_window[0],
        unobserved_intervals_list=unobserved_intervals_list,
        progress_bar=progress_bar,
    )
    ax.set_ylabel("unit #")
    if len(data) <= 30:
        unit_id_display = np.array(units.id.data[:])[[x for x in this_iter]]
        ax.set_yticklabels(unit_id_display)
    else:
        ax.axes.yaxis.set_visible(False)

    return ax


class RasterWidget(widgets.HBox):
    def __init__(
            self,
            units: Units,
            foreign_time_window_controller: StartAndDurationController = None,
            foreign_group_and_sort_controller: GroupAndSortController = None,
            group_by=None,
    ):
        super().__init__()

        self.units = units

        if foreign_time_window_controller is None:
            self.tmin = get_min_spike_time(units)
            self.tmax = get_max_spike_time(units)
            self.time_window_controller = StartAndDurationController(
                tmin=self.tmin, tmax=self.tmax
            )
        else:
            self.time_window_controller = foreign_time_window_controller

        if foreign_group_and_sort_controller:
            self.gas = foreign_group_and_sort_controller
        else:
            self.gas = self.make_group_and_sort(group_by=group_by, control_order=False)

        self.progress_bar = widgets.HBox()

        self.controls = dict(time_window=self.time_window_controller, gas=self.gas)

        plot_func = partial(
            show_session_raster, units=self.units, progress_bar=self.progress_bar
        )

        out_fig = interactive_output(plot_func, self.controls)

        if foreign_time_window_controller:
            right_panel = widgets.VBox(
                children=[
                    self.progress_bar,
                    out_fig,
                ],
                layout=Layout(width="100%"),
            )
        else:
            right_panel = widgets.VBox(
                children=[
                    self.time_window_controller,
                    self.progress_bar,
                    out_fig,
                ],
                layout=Layout(width="100%"),
            )

        if foreign_group_and_sort_controller:
            self.children = [right_panel]
        else:
            self.children = [self.gas, right_panel]

        self.layout = Layout(width="100%")

    def make_group_and_sort(self, group_by=None, control_order=True):
        return GroupAndSortController(
            self.units, group_by=group_by, control_order=control_order
        )


def show_decomposition_series(node, **kwargs):
    # Use Rendering... as a placeholder
    ntabs = 2
    children = [widgets.HTML("Rendering...") for _ in range(ntabs)]

    def on_selected_index(change):
        # Click on Traces Tab
        if change.new == 1 and isinstance(change.owner.children[1], widgets.HTML):
            widget_box = show_decomposition_traces(node)
            children[1] = widget_box
            change.owner.children = children

    field_lay = widgets.Layout(
        max_height="40px", max_width="500px", min_height="30px", min_width="130px"
    )
    vbox = []
    for key, val in node.fields.items():
        lbl_key = widgets.Label(key + ":", layout=field_lay)
        lbl_val = widgets.Label(str(val), layout=field_lay)
        vbox.append(widgets.HBox(children=[lbl_key, lbl_val]))
    children[0] = widgets.VBox(vbox)

    tab_nest = widgets.Tab()
    tab_nest.children = children
    tab_nest.set_title(0, "Fields")
    tab_nest.set_title(1, "Traces")
    tab_nest.observe(on_selected_index, names="selected_index")
    return tab_nest


def show_decomposition_traces(node: DecompositionSeries):
    # Produce figure
    def control_plot(x0, x1, ch0, ch1):
        fig, ax = plt.subplots(nrows=nBands, ncols=1, sharex=True, figsize=(14, 7))
        for bd in range(nBands):
            data = node.data[x0:x1, ch0: ch1 + 1, bd]
            xx = np.arange(x0, x1)
            mu_array = np.mean(data, 0)
            sd_array = np.std(data, 0)
            offset = np.mean(sd_array) * 5
            yticks = [i * offset for i in range(ch1 + 1 - ch0)]
            for i in range(ch1 + 1 - ch0):
                ax[bd].plot(xx, data[:, i] - mu_array[i] + yticks[i])
            ax[bd].set_ylabel("Ch #", fontsize=20)
            ax[bd].set_yticks(yticks)
            ax[bd].set_yticklabels([str(i) for i in range(ch0, ch1 + 1)])
            ax[bd].tick_params(axis="both", which="major", labelsize=16)
        ax[bd].set_xlabel("Time [ms]", fontsize=20)
        return fig

    nSamples = node.data.shape[0]
    nChannels = node.data.shape[1]
    nBands = node.data.shape[2]
    fs = node.rate

    # Controls
    field_lay = widgets.Layout(
        max_height="40px", max_width="100px", min_height="30px", min_width="70px"
    )
    x0 = widgets.BoundedIntText(
        value=0, min=0, max=int(1000 * nSamples / fs - 100), layout=field_lay
    )
    x1 = widgets.BoundedIntText(
        value=nSamples, min=100, max=int(1000 * nSamples / fs), layout=field_lay
    )
    ch0 = widgets.BoundedIntText(
        value=0, min=0, max=int(nChannels - 1), layout=field_lay
    )
    ch1 = widgets.BoundedIntText(
        value=10, min=0, max=int(nChannels - 1), layout=field_lay
    )

    controls = {"x0": x0, "x1": x1, "ch0": ch0, "ch1": ch1}
    out_fig = widgets.interactive_output(control_plot, controls)

    # Assemble layout box
    lbl_x = widgets.Label("Time [ms]:", layout=field_lay)
    lbl_ch = widgets.Label("Ch #:", layout=field_lay)
    lbl_blank = widgets.Label("    ", layout=field_lay)
    hbox0 = widgets.HBox(children=[lbl_x, x0, x1, lbl_blank, lbl_ch, ch0, ch1])
    vbox = widgets.VBox(children=[hbox0, out_fig])
    return vbox


class PSTHWidget(widgets.VBox):
    def __init__(
            self,
            input_data: Units,
            trials: TimeIntervals = None,
            unit_index=0,
            unit_controller=None,
            ntt=1000,
    ):

        self.units = input_data

        super().__init__()

        if trials is None:
            self.trials = self.get_trials()
            if self.trials is None:
                self.children = [widgets.HTML("No trials present")]
                return
        else:
            self.trials = trials

        if unit_controller is None:
            self.unit_ids = self.units.id.data[:]
            n_units = len(self.unit_ids)
            self.unit_controller = widgets.Dropdown(
                options=[(str(self.unit_ids[x]), x) for x in range(n_units)],
                value=unit_index,
                description="unit",
                layout=Layout(width="200px"),
            )
        else:
            self.unit_controller = unit_controller

        self.trial_event_controller = make_trial_event_controller(
            self.trials, layout=Layout(width="200px"), multiple=True
        )
        self.start_ft = widgets.FloatText(
            -0.5, step=0.1, description="start (s)", layout=Layout(width="200px"),
            description_tooltip='Start time for calculation before or after (negative or positive) the reference point (aligned to)'
        )

        self.end_ft = widgets.FloatText(
            1.0, step=0.1, description="end (s)", layout=Layout(width="200px"),
            description_tooltip='End time for calculation before or after (negative or positive) the reference point (aligned to).'
        )

        self.psth_type_radio = widgets.RadioButtons(
            options=["histogram", "gaussian"], layout=Layout(width="100px")
        )
        self.bins_ft = widgets.IntText(
            30, min=0, description="# bins", layout=Layout(width="150px")
        )
        self.gaussian_sd_ft = widgets.FloatText(
            0.05,
            min=0.001,
            description="sd (s)",
            layout=Layout(width="150px"),
            active=False,
            step=0.01,
        )

        self.gas = self.make_group_and_sort(window=False, control_order=False)

        self.controls = dict(
            ntt=fixed(ntt),
            index=self.unit_controller,
            end=self.end_ft,
            start=self.start_ft,
            start_labels=self.trial_event_controller,
            gas=self.gas,
            plot_type=self.psth_type_radio,
            sigma_in_secs=self.gaussian_sd_ft,
            nbins=self.bins_ft
            # progress_bar=fixed(progress_bar)
        )

        out_fig = interactive_output(self.update, self.controls)

        self.children = [
            widgets.HBox(
                [
                    widgets.VBox(
                        [
                            self.gas,
                            widgets.HBox(
                                [
                                    self.psth_type_radio,
                                    widgets.VBox([self.gaussian_sd_ft, self.bins_ft]),
                                ]
                            ),
                        ]
                    ),
                    widgets.VBox(
                        [
                            self.unit_controller,
                            self.trial_event_controller,
                            self.start_ft,
                            self.end_ft,
                        ]
                    ),
                ]
            ),
            out_fig,
        ]

    def get_trials(self):
        return self.units.get_ancestor("NWBFile").trials

    def make_group_and_sort(self, window=None, control_order=False):
        return GroupAndSortController(
            self.trials, window=window, control_order=control_order
        )

    def update(
            self,
            index: int,
            start_labels: tuple = ("start_time",),
            start: float = 0.0,
            end: float = 1.0,
            order=None,
            group_inds=None,
            labels=None,
            sigma_in_secs=0.05,
            ntt: int = 1000,
            progress_bar=None,
            figsize=(12, 7),
            nbins=30,
            plot_type="histogram",
            align_line_color=(0.7, 0.7, 0.7),
    ):
        """

        Parameters
        ----------
        index: int
            Index of unit
        start_label: str, optional
            Trial column name to align on
        start: float
            Start time for calculation before or after (negative or positive) the reference point (aligned to).
        end: float
            End time for calculation before or after (negative or positive) the reference point (aligned to).
        order
        group_inds
        labels
        sigma_in_secs: float, optional
            standard deviation of gaussian kernel
        ntt:
            Number of time points to use for smooth curve
        progress_bar:
        figsize: tuple, optional

        Returns
        -------
        matplotlib.Figure

        """
        fig, axs = plt.subplots(2, len(start_labels), figsize=figsize,
                                sharex=True)
        clean_axes(axs.ravel())

        ax1_ylims = []
        for i_s, start_label in enumerate(start_labels):
            if len(start_labels) > 1:
                ax0 = axs[0, i_s]
                ax1 = axs[1, i_s]
            else:
                ax0 = axs[0]
                ax1 = axs[1]

            data = align_by_time_intervals(
                self.units,
                index,
                self.trials,
                start_label,
                start_label,
                start,
                end,
                order,
                progress_bar=progress_bar,
            )

            if i_s == len(start_labels) - 1:
                show_legend = True
            else:
                show_legend = False

            show_psth_raster(
                data,
                start,
                end,
                group_inds,
                labels,
                show_legend=show_legend,
                ax=ax0,
                progress_bar=progress_bar,
                fontsize=12,
            )

            ax0.set_title(f"{start_label}")
            ax0.set_xticks([])
            ax0.set_xlabel("")

            if i_s > 0:
                ax0.set_ylabel("")
                # Raster always show the same number of trials. We can avoid showing tick labels too
                ax0.set_yticklabels([])

            if plot_type == "gaussian":
                self.bins_ft.layout.visibility = "hidden"
                self.bins_ft.layout.height = "0px"
                self.gaussian_sd_ft.layout.visibility = None
                self.gaussian_sd_ft.layout.height = None
                # expanded data so that gaussian smoother uses larger window than is viewed
                expanded_data = align_by_time_intervals(
                    units=self.units,
                    index=index,
                    intervals=self.trials,
                    start_label=start_label,
                    stop_label=start_label,
                    start=start - sigma_in_secs * 4,
                    end=end + sigma_in_secs * 4,
                    rows_select=order,
                    progress_bar=progress_bar,
                )
                show_psth_smoothed(
                    data=expanded_data,
                    ax=ax1,
                    start=start - sigma_in_secs * 4,
                    end=end + sigma_in_secs * 4,
                    group_inds=group_inds,
                    sigma_in_secs=sigma_in_secs,
                    ntt=ntt,
                )
            elif plot_type == "histogram":
                self.gaussian_sd_ft.layout.visibility = "hidden"
                self.gaussian_sd_ft.layout.height = "0px"
                self.bins_ft.layout.visibility = None
                self.bins_ft.layout.height = None
                show_histogram(data, ax1, start, end, group_inds, nbins=nbins)
            else:
                raise ValueError(
                    "unsupported plot type {}".format(self.psth_type_radio.value)
                )

            ax1.set_xlim([start, end])
            if i_s == 0:
                ax1.set_ylabel("firing rate (Hz)", fontsize=12)
            ax1.set_xlabel("time (s)", fontsize=12)
            ax1.axvline(color=align_line_color)
            ax1_ylims.append(ax1.get_ylim())

        if len(start_labels) > 1:
            # Adjust bottom axes y axis
            min_y = np.min(np.array(ax1_ylims)[:, 0])
            max_y = np.max(np.array(ax1_ylims)[:, 1])
            for i_b, ax_btm in enumerate(axs[1, :]):
                ax_btm.set_ylim(min_y, max_y)
                if i_b > 0:
                    ax_btm.set_ylabel("")
                    # After adjusting ylims we can avoid showing tick labels
                    ax_btm.set_yticklabels([])

        fig.suptitle(f"Unit {self.unit_ids[index]}", fontsize=15)
        fig.subplots_adjust(wspace=0.3)
        return fig


def show_histogram(
        data, ax: plt.Axes, start: float, end: float, group_inds=None, nbins: int = 30
):
    if not len(data):
        return

    if group_inds is None:
        height, x = np.histogram(np.hstack(data), bins=nbins, range=(start, end))
        width = np.diff(x[:2])
        height = height / len(data) / width
        ax.bar(x[:-1], height, edgecolor=(0.3, 0.3, 0.3), width=width, align="edge")
    else:
        data = np.asarray(data, dtype="object")
        # group_inds = np.asarray(group_inds)
        for group in np.unique(group_inds):
            this_data = np.hstack(data[group_inds == group])
            height, x = np.histogram(this_data, bins=nbins, range=(start, end))
            width = np.diff(x[:2])
            height = height / np.sum(group_inds == group) / width
            ax.bar(
                x[:-1],
                height,
                color=color_wheel[group % len(color_wheel)],
                edgecolor=(0.3, 0.3, 0.3),
                width=width,
                align="edge",
                alpha=0.6,
            )


def show_psth_smoothed(
        data,
        ax,
        start: float,
        end: float,
        group_inds=None,
        sigma_in_secs: float = 0.05,
        ntt: int = 1000,
):
    if not len(data):  # TODO: when does this occur?
        return
    all_data = np.hstack(data)
    if not len(all_data):  # no spikes
        return
    tt = np.linspace(start, end, ntt)
    smoothed = np.array(
        [compute_smoothed_firing_rate(x, tt, sigma_in_secs) for x in data]
    )

    if group_inds is None:
        group_inds = np.zeros((len(smoothed)), dtype=int)
    group_stats = []
    for group in np.unique(group_inds):
        this_mean = np.mean(smoothed[group_inds == group], axis=0)
        err = scipy.stats.sem(smoothed[group_inds == group], axis=0)
        group_stats.append(
            dict(
                mean=this_mean,
                lower=this_mean - 2 * err,
                upper=this_mean + 2 * err,
                group=group,
            )
        )
    for stats in group_stats:
        color = color_wheel[stats["group"] % len(color_wheel)]
        ax.plot(tt, stats["mean"], color=color)
        ax.fill_between(tt, stats["lower"], stats["upper"], alpha=0.2, color=color)


def plot_grouped_events(
        data,
        window,
        group_inds=None,
        colors=color_wheel,
        ax=None,
        labels=None,
        show_legend=True,
        offset=0,
        unobserved_intervals_list=None,
        progress_bar=None,
        figsize=(8, 6),
        fontsize=12,
):
    """

    Parameters
    ----------
    data: array-like
    window: array-like [float, float]
        Time in seconds
    group_inds: array-like dtype=int, optional
    colors: array-like, optional
    ax: plt.Axes, optional
    labels: array-like dtype=str, optional
    show_legend: bool, optional
    offset: number, optional
    unobserved_intervals_list: array-like, optional
    progress_bar: FloatProgress, optional
    figsize: tuple, optional
    fontsize: int, optional

    Returns
    -------

    """

    data = np.asarray(data, dtype="object")
    legend_kwargs = dict()
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
        if hasattr(fig, "canvas"):
            fig.canvas.header_visible = False
        else:
            legend_kwargs.update(bbox_to_anchor=(1.01, 1))
    if group_inds is not None:
        ugroup_inds = np.unique(group_inds)
        handles = []

        if progress_bar is not None:
            this_iter = ProgressBar(
                enumerate(ugroup_inds),
                desc="plotting spikes",
                leave=False,
                total=len(ugroup_inds),
            )
            progress_bar = this_iter.container
        else:
            this_iter = enumerate(ugroup_inds)

        for i, ui in this_iter:
            color = colors[ugroup_inds[i] % len(colors)]
            lineoffsets = np.where(group_inds == ui)[0] + offset
            event_collection = ax.eventplot(
                data[group_inds == ui],
                orientation="horizontal",
                lineoffsets=lineoffsets,
                color=color,
            )
            handles.append(event_collection[0])
        if show_legend:
            ax.legend(
                handles=handles[::-1],
                labels=list(labels[ugroup_inds][::-1]),
                loc="upper left",
                bbox_to_anchor=(1.01, 1),
                **legend_kwargs,
            )
    else:
        ax.eventplot(
            data,
            orientation="horizontal",
            color="k",
            lineoffsets=np.arange(len(data)) + offset,
        )

    if unobserved_intervals_list is not None:
        plot_unobserved_intervals(unobserved_intervals_list, ax, offset=offset)

    ax.set_xlim(window)
    ax.set_xlabel("time (s)", fontsize=fontsize)
    ax.set_ylim(np.array([-0.5, len(data) - 0.5]) + offset)
    if len(data) <= 30:
        ax.set_yticks(range(offset, len(data) + offset))
        ax.set_yticklabels(range(offset, len(data) + offset))

    return ax


def plot_unobserved_intervals(
        unobserved_intervals_list, ax, offset=0, color=(0.85, 0.85, 0.85)
):
    for irow, unobs_intervals in enumerate(unobserved_intervals_list):
        rects = [
            Rectangle(
                (i_interval[0], irow - 0.5 + offset), i_interval[1] - i_interval[0], 1
            )
            for i_interval in unobs_intervals
        ]
        pc = PatchCollection(rects, color=color)
        ax.add_collection(pc)


def show_psth_raster(
        data,
        start=-0.5,
        end=2.0,
        group_inds=None,
        labels=None,
        ax=None,
        show_legend=True,
        align_line_color=(0.7, 0.7, 0.7),
        progress_bar: FloatProgress = None,
        fontsize=12,
) -> plt.Axes:
    """

    Parameters
    ----------
    data: array-like
    start: float
        Start time for calculation before or after (negative or positive) the reference point (aligned to).
    end: float
        End time for calculation before or after (negative or positive) the reference point (aligned to).
    group_inds: array-like, optional
    labels: array-like, optional
    ax: plt.Axes, optional
    show_legend: bool, optional
    align_line_color: array-like, optional
        [R, G, B] (0-1)
        Default = [0.7, 0.7, 0.7]
    progress_bar: FloatProgress, optional
    fontsize: int, optional

    Returns
    -------
    plt.Axes

    """
    if not len(data):
        return ax
    ax = plot_grouped_events(
        data,
        [start, end],
        group_inds,
        color_wheel,
        ax,
        labels,
        show_legend=show_legend,
        progress_bar=progress_bar,
        fontsize=fontsize
    )
    ax.set_ylabel("trials", fontsize=fontsize)
    ax.axvline(color=align_line_color)
    return ax


def raster_grid(
        units: Units,
        time_intervals: TimeIntervals,
        index,
        start,
        end,
        rows_label=None,
        cols_label=None,
        trials_select=None,
        align_by="start_time",
) -> plt.Figure:
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    time_intervals: pynwb.epoch.TimeIntervals
    index: int
    start: float
        Start time for calculation before or after (negative or positive) the reference point (aligned to).
    end: float
        End time for calculation before or after (negative or positive) the reference point (aligned to).
    rows_label: str, optional
    cols_label: str, optional
    trials_select: np.array(dtype=bool), optional
    align_by: str, optional

    Returns
    -------
    plt.Figure

    """
    if time_intervals is None:
        raise ValueError("trials must exist (trials cannot be None)")

    if trials_select is None:
        trials_select = np.ones((len(time_intervals),)).astype("bool")

    if rows_label is not None:
        row_vals = np.array(time_intervals[rows_label][:])
        urow_vals = np.unique(row_vals[trials_select])
        if urow_vals.dtype == np.float64:
            urow_vals = urow_vals[~np.isnan(urow_vals)]

    else:
        urow_vals = [None]
    nrows = len(urow_vals)

    if cols_label is not None:
        col_vals = np.array(time_intervals[cols_label][:])
        ucol_vals = np.unique(col_vals[trials_select])
        if ucol_vals.dtype == np.float64:
            ucol_vals = ucol_vals[~np.isnan(ucol_vals)]

    else:
        ucol_vals = [None]
    ncols = len(ucol_vals)

    fig, axs = plt.subplots(
        nrows, ncols, sharex=True, sharey=True, squeeze=False, figsize=(10, 10)
    )
    big_ax = create_big_ax(fig)

    for i, row in enumerate(urow_vals):
        for j, col in enumerate(ucol_vals):
            ax = axs[i, j]
            ax_trials_select = trials_select.copy()
            if row is not None:
                ax_trials_select &= row_vals == row
            if col is not None:
                ax_trials_select &= col_vals == col
            ax_trials_select = np.where(ax_trials_select)[0]
            if len(ax_trials_select):
                data = align_by_time_intervals(
                    units,
                    index,
                    time_intervals,
                    align_by,
                    align_by,
                    start,
                    end,
                    ax_trials_select,
                )
                show_psth_raster(data, start, end, ax=ax)
                ax.set_xlabel("")
                ax.set_ylabel("")
                if ax.get_subplotspec().is_first_col():
                    ax.set_ylabel(row)
                if ax.get_subplotspec().is_last_row():
                    ax.set_xlabel(col)

    big_ax.set_xlabel(cols_label, labelpad=50)
    big_ax.set_ylabel(rows_label, labelpad=60)

    return fig


def plot_grouped_events_plotly(
        data,
        window=None,
        group_inds=None,
        colors=color_wheel,
        labels=None,
        show_legend=True,
        unobserved_intervals_list=None,
        progress_bar=None,
        fig=None,
        **kwargs,
):
    data = np.array(data, dtype=object)

    if fig is None:
        fig = go.FigureWidget()
    if group_inds is not None:
        ugroup_inds = np.unique(group_inds)
        offset = 0
        for i in np.arange(len(ugroup_inds)):
            ui = ugroup_inds[i]
            color = colors[ugroup_inds[i] % len(colors)]
            this_data = data[group_inds == ui]
            event_group(
                this_data,
                offset=offset,
                label=labels[ui],
                color=color,
                fig=fig,
                **kwargs,
            )
            offset += len(this_data)

    else:
        event_group(data, fig=fig, **kwargs)

    fig.update_layout(xaxis_title="time (s)")

    return fig


class RasterWidgetPlotly(widgets.HBox):
    def __init__(
            self,
            units: Units,
            foreign_time_window_controller: StartAndDurationController = None,
            foreign_group_and_sort_controller: GroupAndSortController = None,
            group_by=None,
            fig: go.FigureWidget = None,
    ):
        super().__init__()

        self.units = units

        if foreign_time_window_controller is None:
            self.tmin = get_min_spike_time(units)
            self.tmax = get_max_spike_time(units)
            self.time_window_controller = StartAndDurationController(
                tmin=self.tmin, tmax=self.tmax
            )
        else:
            self.time_window_controller = foreign_time_window_controller

        if foreign_group_and_sort_controller:
            self.gas = foreign_group_and_sort_controller
        else:
            self.gas = GroupAndSortController(
                dynamic_table=self.units, group_by=group_by
            )

        self.show_legend_cb = widgets.Checkbox(value=True, description="show legend")

        if fig is None:
            self.fig = go.FigureWidget()
            self.fig.update_layout(margin=dict(l=20, r=20, t=30, b=20))
        else:
            self.fig = fig
        show_session_raster_plotly(
            self.units, self.fig, self.time_window_controller.value, **self.gas.value
        )

        # set children
        if foreign_time_window_controller:
            right_panel = widgets.VBox(
                children=[
                    self.progress_bar,
                    self.fig,
                ],
                layout=Layout(width="100%"),
            )
        else:
            right_panel = widgets.VBox(
                children=[self.time_window_controller, self.fig, self.show_legend_cb],
                layout=Layout(width="100%"),
            )

        if foreign_group_and_sort_controller:
            self.children = [right_panel]
        else:
            self.children = [self.gas, right_panel]

        self.layout = Layout(width="100%")

        self.time_window_controller.observe(self.update_fig, "value")
        self.gas.observe(self.update_fig, "value")
        self.show_legend_cb.observe(self.toggle_legend, "value")

    def toggle_legend(self, change):
        self.fig.update_layout(showlegend=self.show_legend_cb.value)

    def update_fig(self, change):
        time_window = self.time_window_controller.value
        gas_kwargs = self.gas.value
        with self.fig.batch_update():
            self.fig.data = None
            show_session_raster_plotly(self.units, self.fig, time_window, **gas_kwargs)


def show_session_raster_plotly(
        units: Units, fig, time_window=None, order=None, progress_bar=None, **kwargs
):
    """

    Parameters
    ----------
    units: pynwb.misc.Units
    time_window: [int, int]
    show_obs_intervals: bool
    order: array-like, optional
    group_inds: array-like, optional
    labels: array-like, optional
    show_legend: bool
        default = True
        Does not show legend if color_by is None or 'id'.
    progress_bar: FloatProgress, optional

    Returns
    -------
    go.FigureWidget

    """

    if time_window is None:
        time_window = [get_min_spike_time(units), get_max_spike_time(units)]

    if order is None:
        order = np.arange(len(units), dtype="int")

    if progress_bar:
        this_iter = ProgressBar(order, desc="reading spike data", leave=False)
        progress_bar = this_iter.container
    else:
        this_iter = order
    data = []
    for unit in this_iter:
        data.append(get_spike_times(units, unit, time_window))

    # if show_obs_intervals:
    #    unobserved_intervals_list = get_unobserved_intervals(units, time_window, order)
    # else:
    #    unobserved_intervals_list = None

    fig.update_yaxes(tickvals=[], ticktext=[])
    if len(order) <= 100:
        kwargs.update(marker="line-ns", line_width=2)
    else:
        kwargs.update(line_width=1)
    fig = plot_grouped_events_plotly(data=data, fig=fig, **kwargs)
    if len(order) <= 40:
        fig.update_yaxes(
            tickvals=np.arange(len(order)), ticktext=[str(i) for i in order]
        )

    fig.update_layout(
        title="units",
        xaxis_title="time (s)",
        legend=dict(x=1.0, y=0, traceorder="reversed"),
        xaxis=dict(range=time_window),
        yaxis=dict(range=[-0.5, len(order) + 0.5]),
    )

    return fig


class UnitsAndTrialsControllerWidget(widgets.VBox):
    InnerWidget = None

    def __init__(
            self,
            units: Units,
            trials: TimeIntervals = None,
            unit_index=0,
            **kwargs
    ):
        """
        Creates a UnitsAndTrials controller that controls InnerWidget.

        Parameters
        ----------
        units: pynwb.misc.Units object
        trials: pynwb.epoch.TimeIntervals object
        unit_index: int
        """
        super().__init__()

        self.units = units
        self.kwargs = kwargs

        # Check if there is trials table and create controller
        if trials is None:
            self.trials = self.get_trials()
            if self.trials is None:
                self.children = [widgets.HTML("No trials present")]
                return
        else:
            self.trials = trials

        # Create variables choice dropdowns
        groups = self.get_groups(self.trials)
        self.rows_controller = widgets.Dropdown(
            options=[None] + list(groups),
            description="rows",
            value=None
        )
        self.rows_controller.observe(self.rows_callback, names='value')

        self.cols_controller = widgets.Dropdown(
            options=[None] + list(groups),
            description="cols",
            disabled=True,
        )

        # Unit controller
        unit_ids = self.units.id.data[:]
        n_units = len(unit_ids)
        self.unit_controller = widgets.Dropdown(
            options=[(str(unit_ids[x]), x) for x in range(n_units)],
            value=unit_index,
            description="unit",
        )

        # Trial event controller (align by)
        self.trial_event_controller = make_trial_event_controller(self.trials)

        # Start / End controllers
        self.start_ft = widgets.FloatText(
            -0.5, step=0.1, description="start (s)", layout=Layout(width="200px"),
            description_tooltip='Start time for calculation before or after (negative or positive) the reference point (aligned to)'
        )
        self.end_ft = widgets.FloatText(
            1.0, step=0.1, description="end (s)", layout=Layout(width="200px"),
            description_tooltip='End time for calculation before or after (negative or positive) the reference point (aligned to).'
        )

        self.fixed = dict(
            units=self.units,
            time_intervals=self.trials,
        )

        self.controls = {
            "index": self.unit_controller,
            "start": self.start_ft,
            "end": self.end_ft,
            "align_by": self.trial_event_controller,
            "rows_label": self.rows_controller,
            "cols_label": self.cols_controller,
        }

        self.children = [
            self.unit_controller,
            self.rows_controller,
            self.cols_controller,
            self.trial_event_controller,
            self.start_ft,
            self.end_ft
        ]

    def get_trials(self):
        return self.units.get_ancestor("NWBFile").trials

    def get_groups(self, trials):
        return infer_categorical_columns(dynamic_table=trials)

    def rows_callback(self, change):
        """
        Gets triggered when self.rows_controller changes. Updates other dropdown options.
        """
        if change['new'] is None:
            self.cols_controller.disabled = True
            self.cols_controller.value = None
        else:
            self.cols_controller.disabled = False


class RasterGridWidget(widgets.VBox):
    def __init__(
            self,
            units: Units,
            trials: TimeIntervals = None,
            unit_index=0,
            units_trials_controller=None,
    ):
        super().__init__()

        # Create Units and Trials controller
        if not units_trials_controller:
            units_trials_controller = UnitsAndTrialsControllerWidget(
                units=units,
                trials=trials,
                unit_index=unit_index
            )
            self.children = [units_trials_controller]

        self.fig = interactive_output(
            f=raster_grid,
            controls=units_trials_controller.controls,
            fixed=units_trials_controller.fixed
        )

        self.children += tuple([self.fig])


class TuningCurveWidget(widgets.VBox):
    def __init__(
            self,
            units: Units,
            trials: TimeIntervals = None,
            unit_index=0,
            units_trials_controller=None,
    ):
        super().__init__()
        self.children = []

        # Create Units and Trials controller
        if not units_trials_controller:
            units_trials_controller = UnitsAndTrialsControllerWidget(
                units=units,
                trials=trials,
                unit_index=unit_index
            )
            self.children = [units_trials_controller]

        self.fig = interactive_output(
            f=draw_tuning_curve,
            controls=units_trials_controller.controls,
            fixed=units_trials_controller.fixed
        )

        self.children += tuple([self.fig])


class TuningCurveExtendedWidget(widgets.VBox):
    def __init__(
            self,
            units: Units,
            trials: TimeIntervals = None,
            unit_index=0
    ):
        super().__init__()

        # Controller
        self.units_trials_controller = UnitsAndTrialsControllerWidget(
            units=units,
            trials=trials,
            unit_index=unit_index
        )

        # Tuning curve widget
        self.tuning_curve = TuningCurveWidget(
            units=units,
            trials=trials,
            unit_index=unit_index,
            units_trials_controller=self.units_trials_controller,
        )

        # Raster grid widget
        self.raster_grid = RasterGridWidget(
            units=units,
            trials=trials,
            unit_index=unit_index,
            units_trials_controller=self.units_trials_controller,
        )

        self.children = [
            self.units_trials_controller,
            self.tuning_curve,
            self.raster_grid
        ]


def draw_tuning_curve(
        units: Units,
        time_intervals: TimeIntervals,
        index,
        start,
        end,
        rows_label=None,
        cols_label=None,
        align_by="start_time",
) -> plt.Figure:
    if rows_label is None:
        return widgets.HTML("Select at least one variable")

    # 1D histogram
    if cols_label is None:
        return draw_tuning_curve_1d(
            units,
            time_intervals,
            index,
            start,
            end,
            rows_label,
            align_by
        )

    return draw_tuning_curve_2d(
        units,
        time_intervals,
        index,
        start,
        end,
        rows_label,
        cols_label,
        align_by
    )


def draw_tuning_curve_1d(
        units: Units,
        time_intervals: TimeIntervals,
        index,
        start,
        end,
        rows_label=None,
        align_by="start_time",
) -> plt.Figure:
    rows_data, var1_classes = extract_data_from_intervals(time_intervals[rows_label])

    avg_rates = []
    for v1 in var1_classes:
        indexes = [i for i, d in enumerate(rows_data) if d == v1]
        data = align_by_time_intervals(
            units=units,
            index=index,
            intervals=time_intervals,
            start_label=align_by,
            stop_label=align_by,
            start=start,
            end=end,
            rows_select=indexes
        )
        n_trials = len(data)
        n_spikes = len(np.hstack(data))
        duration = end - start
        avg_rates.append(n_spikes / (n_trials * duration))

    x = np.arange(len(var1_classes))  # the label locations
    si = sort_mixed_type_list(var1_classes)
    avg_rates_sorted = [avg_rates[i] for i in si]

    fig, ax = plt.subplots(figsize=(14, 7))
    width = 0.95  # the width of the bars
    rects1 = ax.bar(x, avg_rates_sorted, width)
    lines1 = ax.plot(x, avg_rates_sorted, '-o', color='k', lw=2)

    # Labels
    ax.set_ylabel('Avg rate')
    ax.set_xlabel(rows_label)
    ax.set_xticks(x)
    ax.set_xticklabels([var1_classes[i] for i in si], rotation=45)
    fig.tight_layout()


def draw_tuning_curve_2d(
        units: Units,
        time_intervals: TimeIntervals,
        index,
        start,
        end,
        rows_label=None,
        cols_label=None,
        align_by="start_time",
) -> plt.Figure:
    rows_data, var1_classes = extract_data_from_intervals(time_intervals[rows_label])
    cols_data, var2_classes = extract_data_from_intervals(time_intervals[cols_label])

    avg_rates = np.zeros((len(var1_classes), len(var2_classes)))
    for i, v1 in enumerate(var1_classes):
        for j, v2 in enumerate(var2_classes):
            indexes1 = [ii for ii, d in enumerate(rows_data) if d == v1]
            indexes2 = [ii for ii, d in enumerate(cols_data) if d == v2]
            intersect = list(set(indexes1) & set(indexes2))
            if len(intersect) > 0:
                data = align_by_time_intervals(
                    units=units,
                    index=index,
                    intervals=time_intervals,
                    start_label=align_by,
                    stop_label=align_by,
                    start=start,
                    end=end,
                    rows_select=intersect
                )
                n_trials = len(data)
                n_spikes = len(np.hstack(data))
                duration = end - start
                avg_rates[i, j] = n_spikes / (n_trials * duration)

    fig, ax = plt.subplots(figsize=(14, 7))
    pos = ax.imshow(avg_rates.T, origin='lower', cmap='Greys')
    cbar = fig.colorbar(pos, ax=ax)
    cbar.set_label('spikes / second')

    # Labels
    ax.set_xticks(np.arange(len(var1_classes)))
    ax.set_yticks(np.arange(len(var2_classes)))
    ax.set_xlabel(rows_label)
    ax.set_ylabel(cols_label)
    ax.set_xticklabels(var1_classes, rotation=45)
    ax.set_yticklabels(var2_classes, rotation=45)

    return fig


def sort_mixed_type_list(x):
    """Returns the indexes for a sorted list of mixed types"""
    x_num = list()
    x_num_i = list()
    x_oth = list()
    x_oth_i = list()
    for i, xx in enumerate(x):
        try:
            x_num.append(float(xx))
            x_num_i.append(i)
        except:
            x_oth.append(str(xx))
            x_oth_i.append(i)
    x_num_si = np.argsort(x_num)
    x_oth_si = np.argsort(x_oth)
    return [x_num_i[ii] for ii in x_num_si] + [x_oth_i[ii] for ii in x_oth_si]



import numpy as np
import matplotlib.pyplot as plt

import ipywidgets as widgets

from pynwb.base import TimeSeries
from pynwb.image import RGBImage, GrayscaleImage, IndexSeries, ImageSeries

from nwbwidgets.image import (
    show_rbga_image,
    show_grayscale_image,
    show_index_series,
    show_image_series,
)
from nwbwidgets.view import default_neurodata_vis_spec


def test_show_rbg_image():
    data = np.random.rand(2700).reshape((30, 30, 3))
    rgb_image = RGBImage(name="test_image", data=data)

    assert isinstance(show_rbga_image(rgb_image), plt.Figure)


def test_show_grayscale_image():
    data = np.random.rand(900).reshape((30, 30))
    grayscale_image = GrayscaleImage(name="test_image", data=data)

    assert isinstance(show_grayscale_image(grayscale_image), plt.Figure)


def test_show_index_series():
    data = np.array([12, 14, 16, 18, 20, 22, 24, 26])
    indexed_timeseries = TimeSeries(
        name="Index Series time data",
        data=np.random.rand(800).reshape((8, 10, 10)),
        rate=1.0,
        unit='na',
    )
    index_series = IndexSeries(
        name="Sample Index Series",
        data=data,
        indexed_timeseries=indexed_timeseries,
        rate=1.0,
        unit='n.a.',
    )

    assert isinstance(
        show_index_series(index_series, default_neurodata_vis_spec), widgets.Widget
    )


def test_show_image_series():
    data = np.random.rand(800).reshape((8, 10, 10))
    image_series = ImageSeries(name="Image Series", data=data, rate=1.0, unit='n.a.')

    assert isinstance(
        show_image_series(image_series, default_neurodata_vis_spec), widgets.Widget
    )



import plotly.io as pio

from .view import nwb2widget, default_neurodata_vis_spec

# from .ephys_viz_interface import ephys_viz_neurodata_vis_spec

from .version import version as __version__

pio.templates.default = "simple_white"



import unittest

import numpy as np

from hdmf.common import DynamicTable, VectorData
from pynwb.ecephys import ElectrodeGroup, Device

from nwbwidgets.controllers import (RangeController, GroupAndSortController,
                                    StartAndDurationController)


class FloatRangeControllerTestCase(unittest.TestCase):
    def setUp(self):
        self.range_controller = RangeController(vmin=0, vmax=10, start_value=(5, 7))

    def test_move_range_slider_down_bigger(self):
        self.range_controller.value = (4, 6)
        self.range_controller.move_down("filler")
        assert self.range_controller.value == (2, 4)

    def test_move_range_slider_down_smaller(self):
        self.range_controller.value = (2, 6)
        self.range_controller.move_down("filler")
        assert self.range_controller.value == (0, 4)

    def test_move_range_slider_up_smaller(self):
        self.range_controller.value = (5, 7)
        self.range_controller.move_up("filler")
        assert self.range_controller.value == (7, 9)

    def test_move_range_slider_up_bigger(self):
        self.range_controller.value = (5, 8)
        self.range_controller.move_up("filler")
        assert self.range_controller.value == (7, 10)


class TestGroupAndSortController(unittest.TestCase):
    def setUp(self) -> None:
        data1 = np.array([1, 2, 2, 3, 1, 1, 3, 2, 3])
        data2 = np.array([3, 4, 2, 4, 3, 2, 2, 4, 4])
        device = Device(name="device")
        eg_1 = ElectrodeGroup(
            name="electrodegroup1", description="desc", location="brain", device=device
        )
        eg_2 = ElectrodeGroup(
            name="electrodegroup2", description="desc", location="brain", device=device
        )
        data3 = [eg_1, eg_2, eg_1, eg_1, eg_1, eg_1, eg_1, eg_1, eg_1]
        vd1 = VectorData("Data1", "vector data for creating a DynamicTable", data=data1)
        vd2 = VectorData("Data2", "vector data for creating a DynamicTable", data=data2)
        vd3 = VectorData(
            "ElectrodeGroup", "vector data for creating a DynamicTable", data=data3
        )
        vd = [vd1, vd2, vd3]

        self.dynamic_table = DynamicTable(
            name="test table",
            description="This is a test table",
            columns=vd,
            colnames=["Data1", "Data2", "ElectrodeGroup"],
        )

    def test_all_rows(self):
        GroupAndSortController(dynamic_table=self.dynamic_table)

    def test_keep_rows(self):
        keep_rows = np.arange(len(self.dynamic_table) // 2)
        GroupAndSortController(dynamic_table=self.dynamic_table, keep_rows=keep_rows)

    def test_control(self):
        gas = GroupAndSortController(dynamic_table=self.dynamic_table)

        gas.group_dd.value = "Data1"
        gas.group_dd.value = None

        gas.order_dd.value = "Data1"
        gas.order_dd.value = None

class TestStartAndDurationController(unittest.TestCase):
    def setUp(self) -> None:
        self.start_and_duration_controller = StartAndDurationController(10)

    def test_set_duration(self):
        self.start_and_duration_controller.duration.value = 2

    def test_set_start(self):
        self.start_and_duration_controller.slider.value = 4

    def test_set_start_against_max(self):
        self.start_and_duration_controller.slider.value = 9
        assert self.start_and_duration_controller.slider.value == 5

    def test_buttons(self):
        self.start_and_duration_controller.to_end_button.click()
        self.start_and_duration_controller.to_start_button.click()
        self.start_and_duration_controller.forward_button.click()
        self.start_and_duration_controller.backwards_button.click()



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np

import ipywidgets as widgets

from pynwb.device import Device
from pynwb.ophys import (
    TwoPhotonSeries,
    OpticalChannel,
    ImageSegmentation,
    Fluorescence,
    DfOverF,
    ImagingPlane,
    PlaneSegmentation
)

from ndx_grayscalevolume import GrayscaleVolume

from nwbwidgets.ophys import (
    TwoPhotonSeriesWidget,
    show_grayscale_volume,
    show_df_over_f,
    PlaneSegmentation2DWidget,
    show_plane_segmentation_3d_voxel,
    show_plane_segmentation_3d_mask,
    show_image_segmentation,
)
from nwbwidgets.view import default_neurodata_vis_spec


def test_show_grayscale_volume():
    vol = GrayscaleVolume(name="vol", data=np.random.rand(2700).reshape((30, 30, 3)))
    assert isinstance(
        show_grayscale_volume(vol, default_neurodata_vis_spec), widgets.Widget
    )


class CalciumImagingTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        device = Device("imaging_device_1")
        optical_channel = OpticalChannel("my_optchan", "description", 500.0)
        self.imaging_plane = ImagingPlane(
            name="imgpln1",
            optical_channel=optical_channel,
            description="a fake ImagingPlane",
            device=device,
            excitation_lambda=600.0,
            imaging_rate=300.0,
            indicator="GFP",
            location="somewhere in the brain",
            reference_frame="unknown",
            origin_coords=[10, 20],
            origin_coords_unit="millimeters",
            grid_spacing=[0.001, 0.001],
            grid_spacing_unit="millimeters",
        )

        self.image_series = TwoPhotonSeries(
            name="test_image_series",
            data=np.random.randn(100, 5, 5),
            imaging_plane=self.imaging_plane,
            starting_frame=[0],
            rate=1.0,
            unit="n.a",
        )
        self.img_seg = ImageSegmentation()
        self.ps2 = self.img_seg.create_plane_segmentation(
            "output from segmenting my favorite imaging plane",
            self.imaging_plane,
            "2d_plane_seg",
            self.image_series,
        )
        self.ps2.add_column("type", "desc")
        self.ps2.add_column("type2", "desc")

        w, h = 3, 3
        img_mask1 = np.zeros((w, h))
        img_mask1[0, 0] = 1.1
        img_mask1[1, 1] = 1.2
        img_mask1[2, 2] = 1.3
        self.ps2.add_roi(image_mask=img_mask1, type=1, type2=0)

        img_mask2 = np.zeros((w, h))
        img_mask2[0, 0] = 2.1
        img_mask2[1, 1] = 2.2
        self.ps2.add_roi(image_mask=img_mask2, type=1, type2=1)

        img_mask2 = np.zeros((w, h))
        img_mask2[0, 0] = 9.1
        img_mask2[1, 1] = 10.2
        self.ps2.add_roi(image_mask=img_mask2, type=2, type2=0)

        img_mask2 = np.zeros((w, h))
        img_mask2[0, 0] = 3.5
        img_mask2[1, 1] = 5.6
        self.ps2.add_roi(image_mask=img_mask2, type=2, type2=1)

        fl = Fluorescence()
        rt_region = self.ps2.create_roi_table_region(
            "the first of two ROIs", region=[0, 1, 2, 3]
        )

        rois_shape = 5
        data = np.arange(10*rois_shape).reshape([10, -1], order='F')
        timestamps = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
        rrs = fl.create_roi_response_series(
            name="my_rrs", data=data, rois=rt_region, unit="lumens", timestamps=timestamps
        )
        self.df_over_f = DfOverF(rrs)

    def test_show_two_photon_series(self):
        wid = TwoPhotonSeriesWidget(self.image_series, default_neurodata_vis_spec)
        assert isinstance(wid, widgets.Widget)
        wid.controls['slider'].value = 50

    def test_show_3d_two_photon_series(self):
        image_series3 = TwoPhotonSeries(
            name="test_3d_images",
            data=np.random.randn(100, 5, 5, 5),
            imaging_plane=self.imaging_plane,
            starting_frame=[0],
            rate=1.0,
            unit="n.a",
        )
        wid = TwoPhotonSeriesWidget(image_series3, default_neurodata_vis_spec)
        assert isinstance(wid, widgets.Widget)
        wid.controls['slider'].value = 50

    def test_show_df_over_f(self):
        dff = show_df_over_f(self.df_over_f, default_neurodata_vis_spec)
        assert isinstance(dff, widgets.Widget)
        dff.controls['gas'].window = [1, 2]

    def test_plane_segmentation_2d_widget(self):
        wid = PlaneSegmentation2DWidget(self.ps2)
        assert isinstance(wid, widgets.Widget)
        wid.button.click()
        wid.cat_controller.value = "type"
        wid.cat_controller.value = "type2"

    def test_show_plane_segmentation_3d_mask(self):
        ps3 = PlaneSegmentation(
            "output from segmenting my favorite imaging plane",
            self.imaging_plane,
            "3d_plane_seg",
            self.image_series,
        )

        w, h, d = 3, 3, 3
        img_mask1 = np.zeros((w, h, d))
        for i in range(3):
            img_mask1[i, i, i] = 1.0
        ps3.add_roi(image_mask=img_mask1)

        img_mask2 = np.zeros((w, h, d))
        for i in range(3):
            img_mask2[i, i, i] = 1.2
        ps3.add_roi(image_mask=img_mask2)
        wid = show_plane_segmentation_3d_mask(ps3)
        assert isinstance(wid, widgets.Widget)

    def test_show_plane_segmentation_3d_voxel(self):
        ps3v = PlaneSegmentation(
            "output from segmenting my favorite imaging plane",
            self.imaging_plane,
            "3d_voxel",
            self.image_series,
        )

        voxel_mask = [(i, i, i, 1.0) for i in range(3)]
        ps3v.add_roi(voxel_mask=voxel_mask)

        voxel_mask = [(1, 1, i, 1.2) for i in range(3)]
        ps3v.add_roi(voxel_mask=voxel_mask)
        wid = show_plane_segmentation_3d_voxel(ps3v)
        assert isinstance(wid, widgets.Widget)

    def test_show_image_segmentation(self):
        assert isinstance(
            show_image_segmentation(self.img_seg, default_neurodata_vis_spec),
            widgets.Widget,
        )



import numpy as np

import ipywidgets as widgets
import plotly.graph_objects as go

from pynwb.base import DynamicTable

import trimesh

from .base import df_to_hover_text


def make_cylinder_mesh(
    radius, height, sections=32, position=(0, 0, 0), direction=(1, 0, 0), **kwargs
):
    new_normal = direction / np.linalg.norm(direction)
    cosx, cosy = new_normal[:2]
    sinx = np.sqrt(1 - cosx ** 2)
    siny = np.sqrt(1 - cosy ** 2)

    yaw = [[cosx, -sinx, 0, 0], [sinx, cosx, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]

    pitch = [[cosy, 0, siny, 0], [0, 1, 0, 0], [-siny, 0, cosy, 0], [0, 0, 0, 1]]

    transform = np.dot(yaw, pitch)

    transform[:3, 3] = position

    cylinder = trimesh.primitives.Cylinder(
        radius=radius, height=height, sections=sections, transform=transform
    )

    x, y, z = cylinder.vertices.T
    i, j, k = cylinder.faces.T

    return go.Mesh3d(x=x, y=y, z=z, i=i, j=j, k=k, **kwargs)


def make_cylinders(
    positions, directions, radius=1, height=1, sections=32, name="cylinders", **kwargs
):

    return [
        make_cylinder_mesh(
            position=position,
            direction=direction,
            radius=radius,
            height=height,
            sections=sections,
            showlegend=not i,
            legendgroup=name,
            name=name,
            **kwargs
        )
        for i, (position, direction) in enumerate(zip(positions, directions))
    ]


class HumanElectrodesPlotlyWidget(widgets.VBox):
    def __init__(self, electrodes: DynamicTable, **kwargs):

        super().__init__()
        self.electrodes = electrodes

        slider_kwargs = dict(
            value=1.0, min=0.0, max=1.0, style={"description_width": "initial"}
        )

        left_opacity_slider = widgets.FloatSlider(
            description="left hemi opacity", **slider_kwargs
        )

        right_opacity_slider = widgets.FloatSlider(
            description="right hemi opacity", **slider_kwargs
        )

        color_by_dropdown = widgets.Dropdown(
            options=list(electrodes.colnames),
            value="group_name",
            description="Color By:",
            disabled=False,
        )

        color_by_dropdown.observe(self.color_electrode_by)
        left_opacity_slider.observe(self.observe_left_opacity)
        right_opacity_slider.observe(self.observe_right_opacity)

        self.fig = go.FigureWidget()
        self.plot_human_brain()
        self.show_electrodes(electrodes, color_by_dropdown.value)
        sliders = widgets.HBox([left_opacity_slider, right_opacity_slider])
        self.children = [self.fig, widgets.VBox([sliders, color_by_dropdown])]

    @staticmethod
    def find_normals(points, k=3):
        normals = []
        for point in points:
            from skspatial.objects import Points, Plane

            distance = np.linalg.norm(points - point, axis=1)
            # closest_inds = np.argpartition(distance, 3)
            # x0, x1, x2 = points[closest_inds[:3]]
            # normal = np.cross((x1 - x0), (x2 - x0))
            closest_inds = np.argpartition(distance, k)
            close_points = points[closest_inds[:k]]
            normal = np.asarray(Plane.best_fit(close_points).normal)
            normals.append(normal)
        return normals

    def show_electrodes(self, electrodes: DynamicTable, color_by):

        positions = np.c_[electrodes.x[:], electrodes.y[:], electrodes.z[:]]

        if isinstance(electrodes[color_by][0], (bytes, str, np.bool_)):
            ugroups, group_inv = np.unique(electrodes[color_by][:], return_inverse=True)
            colors = group_inv
            show_leg = True
            show_scale = False
        elif isinstance(electrodes[color_by][0], (np.ndarray, np.float)):
            colors = np.ravel(electrodes[color_by][:])
            ugroups, group_inv = [0], np.array([0] * len(colors))
            show_leg = False
            show_scale = True

        else:
            print("Not a valid data type")
            return

        c_max = np.max(colors)
        c_min = np.min(colors)

        for i, group in enumerate(ugroups):
            sel_positions = positions[group_inv == i]
            c = colors[group_inv == i]
            x, y, z = sel_positions.T

            if isinstance(group, bytes):
                group = group.decode()

                """
                if 'GRID' in group:
                    normals = self.find_normals(sel_positions, 5)
                    with self.fig.batch_update():
                        [self.fig.add_trace(trace) for trace in make_cylinders(
                            positions=sel_positions,
                            directions=normals,
                            radius=2,
                            height=.5,
                            color=c,
                            name=group
                    )]
                else:


                """
            self.fig.add_trace(
                go.Scatter3d(
                    mode="markers",
                    x=x,
                    y=y,
                    z=z,
                    name=str(group),
                    legendgroup=str(group),
                    marker=dict(
                        color=c,
                        cmax=c_max,
                        cmin=c_min,
                        colorscale="Viridis",
                        colorbar=dict(title="Colorbar"),
                        showscale=show_scale,
                    ),
                    text=df_to_hover_text(electrodes.to_dataframe()),
                    hoverinfo="text",
                    showlegend=show_leg,
                )
            ),

        self.fig.update_layout(
            legend=dict(
                x=0,
                y=1,
            ),
        )

    def plot_human_brain(self, left_opacity=1.0, right_opacity=1.0):

        from nilearn import datasets, surface

        mesh = datasets.fetch_surf_fsaverage("fsaverage5")

        def create_mesh(name, **kwargs):
            vertices, triangles = surface.load_surf_mesh(mesh[name])
            x, y, z = vertices.T
            i, j, k = triangles.T

            return go.Mesh3d(x=x, y=y, z=z, i=i, j=j, k=k, **kwargs)

        kwargs = dict(
            color="lightgray",
            lighting=dict(specular=1, ambient=0.9, roughness=0.9, diffuse=0.9),
            hoverinfo="skip",
        )

        self.fig.add_trace(create_mesh("pial_left", opacity=left_opacity, **kwargs))
        self.fig.add_trace(create_mesh("pial_right", opacity=right_opacity, **kwargs))

        self.fig.update_layout(
            scene=dict(
                xaxis=dict(visible=False),
                yaxis=dict(visible=False),
                zaxis=dict(visible=False),
            ),
            height=500,
            margin=dict(t=20, b=0),
        )

    def observe_left_opacity(self, change):
        if "new" in change and isinstance(change["new"], float):
            self.fig.data[0].opacity = change["new"]

    def observe_right_opacity(self, change):
        if "new" in change and isinstance(change["new"], float):
            self.fig.data[1].opacity = change["new"]

    def color_electrode_by(self, change):
        if "new" in change and isinstance(change["new"], str):
            with self.fig.batch_update():
                self.fig.data = None
                self.plot_human_brain()
                self.show_electrodes(self.electrodes, change["new"])



import numpy as np

import plotly.graph_objects as go


def multi_trace(x, y, color, label=None, fig=None, insert_nans=False):
    """Create multiple traces that are associated with a single legend label

    Parameters
    ----------
    x: array-like
    y: array-like
    color: str
    label: str, optional
    fig: go.FigureWidget

    Returns
    -------

    """
    if fig is None:
        fig = go.FigureWidget()

    if insert_nans:
        y_nans = []
        x_nans = []
        for xx,yy in zip(x,y):
            y_nans.append(np.append(yy,np.nan))
            x_nans.append(np.append(xx, np.nan))
        y_plot = np.concatenate(y_nans,axis=0)
        x_plot = np.concatenate(x_nans, axis=0)
        fig.add_scattergl(
            x=x_plot,
            y=y_plot,
            name=label,
            line={"color": color},
        )
        return fig
    else:
        for i, yy in enumerate(y):
            if label is not None and i:
                showlegend = False
            else:
                showlegend = True

            fig.add_scattergl(
                x=x,
                y=yy,
                legendgroup=label,
                name=label,
                showlegend=showlegend,
                line={"color": color},
            )

        return fig


def event_group(
    times_list,
    offset=0,
    color="Black",
    label=None,
    fig=None,
    marker=None,
    line_width=None,
):
    """Create an event raster that are all associated with a single legend label

    Parameters
    ----------
    times_list: list of array-like
    offset: float, optional
    label: str, optional
    fig: go.FigureWidget

    optional, passed to go.Scatter.marker:
    marker: str
    line_width: str
    color: str
        default: Black


    Returns
    -------

    """
    if fig is None:
        fig = go.FigureWidget()

    if label is not None:
        showlegend = True
    else:
        showlegend = False

    for i, times in enumerate(times_list):
        if len(times):
            fig.add_scattergl(
                x=times,
                y=np.ones_like(times) * (i + offset),
                marker=dict(
                    color=color, line_width=line_width, symbol=marker, line_color=color
                ),
                legendgroup=str(label),
                name=label,
                showlegend=showlegend,
                mode="markers",
            )
            showlegend = False

    return fig



from abc import abstractmethod

import numpy as np
import matplotlib.pyplot as plt

import ipywidgets as widgets
from plotly import graph_objects as go

from pynwb import TimeSeries
from pynwb.behavior import SpatialSeries, BehavioralEvents

from nwbwidgets import base

from .utils.timeseries import (
    get_timeseries_tt,
    get_timeseries_in_units,
    timeseries_time_to_ind,
)
from .timeseries import (
    AlignMultiTraceTimeSeriesByTrialsConstant,
    AlignMultiTraceTimeSeriesByTrialsVariable,
    AbstractTraceWidget,
    SingleTracePlotlyWidget,
    SeparateTracesPlotlyWidget,
)
from .base import lazy_tabs
from .controllers import StartAndDurationController


def show_behavioral_events(beh_events: BehavioralEvents, neurodata_vis_spec: dict):
    return base.dict2accordion(
        beh_events.time_series, neurodata_vis_spec, ls="", marker="|"
    )


def show_spatial_series(node: SpatialSeries, **kwargs):
    data, unit = get_timeseries_in_units(node)
    tt = get_timeseries_tt(node)

    if len(data.shape) == 1:
        fig, ax = plt.subplots()
        ax.plot(tt, data, **kwargs)
        ax.set_xlabel("t (sec)")
        if unit:
            ax.set_xlabel("x ({})".format(unit))
        else:
            ax.set_xlabel("x")
        ax.set_ylabel("x")

    elif data.shape[1] == 2:
        fig, ax = plt.subplots()
        ax.plot(data[:, 0], data[:, 1], **kwargs)
        if unit:
            ax.set_xlabel("x ({})".format(unit))
            ax.set_ylabel("y ({})".format(unit))
        else:
            ax.set_xlabel("x")
            ax.set_ylabel("y")
        ax.axis("equal")

    elif data.shape[1] == 3:
        import ipyvolume.pylab as p3

        fig = p3.figure()
        p3.scatter(data[:, 0], data[:, 1], data[:, 2], **kwargs)
        p3.xlim(np.min(data[:, 0]), np.max(data[:, 0]))
        p3.ylim(np.min(data[:, 1]), np.max(data[:, 1]))
        p3.zlim(np.min(data[:, 2]), np.max(data[:, 2]))

    else:
        raise NotImplementedError

    return fig


def route_spatial_series(spatial_series, **kwargs):
    if len(spatial_series.data.shape) == 1:
        dict_ = {
            "over time": SingleTracePlotlyWidget,
            "trial aligned": trial_align_spatial_series,
        }
    elif spatial_series.data.shape[1] == 2:
        dict_ = {
            "over time": SeparateTracesPlotlyWidget,
            "trace": SpatialSeriesTraceWidget2D,
            "trial aligned": trial_align_spatial_series,
        }
    elif spatial_series.data.shape[1] == 3:
        dict_ = {
            "over time": SeparateTracesPlotlyWidget,
            "trace": SpatialSeriesTraceWidget3D,
            "trial aligned": trial_align_spatial_series,
        }
    else:
        return widgets.HTML("Unsupported number of dimensions.")
    return lazy_tabs(dict_, spatial_series)


class SpatialSeriesTraceWidget(AbstractTraceWidget):
    @abstractmethod
    def plot_data(self, data, units, tt):
        return

    @abstractmethod
    def update_plot(self, data, tt):
        return

    def __init__(
        self,
        timeseries: TimeSeries,
        foreign_time_window_controller: StartAndDurationController = None,
        **kwargs,
    ):
        super().__init__(
            timeseries=timeseries,
            foreign_time_window_controller=foreign_time_window_controller,
            **kwargs,
        )

    def set_out_fig(self):
        timeseries = self.controls["timeseries"].value
        time_window = self.controls["time_window"].value

        istart = timeseries_time_to_ind(timeseries, time_window[0])
        istop = timeseries_time_to_ind(timeseries, time_window[1])
        data, units = get_timeseries_in_units(timeseries, istart, istop)

        tt = get_timeseries_tt(timeseries, istart, istop)

        self.plot_data(data, units, tt)

        def on_change(change):
            time_window = self.controls["time_window"].value
            istart = timeseries_time_to_ind(timeseries, time_window[0])
            istop = timeseries_time_to_ind(timeseries, time_window[1])
            data, units = get_timeseries_in_units(timeseries, istart, istop)
            tt = get_timeseries_tt(timeseries, istart, istop)
            self.update_plot(data, tt)

        self.controls["time_window"].observe(on_change)


class SpatialSeriesTraceWidget2D(SpatialSeriesTraceWidget):
    def plot_data(self, data, units, tt):
        if units is None:
            units = "no units"
        self.out_fig = go.FigureWidget(
            data=go.Scattergl(
                x=list(data[:, 0]),
                y=list(data[:, 1]),
                mode="markers+lines",
                marker_color=tt,
                marker_colorscale="Viridis",
                marker_colorbar=dict(thickness=20, title="time (s)"),
                marker_size=5,
            )
        )

        self.out_fig.update_layout(
            title=self.timeseries.name,
            xaxis_title=f"x ({units})",
            yaxis_title=f"y ({units})",
        )

    def update_plot(self, data, tt):
        self.out_fig.data[0].x = list(data[:, 0])
        self.out_fig.data[0].y = list(data[:, 1])
        self.out_fig.update_traces(marker_color=list(tt))


class SpatialSeriesTraceWidget3D(SpatialSeriesTraceWidget):
    def plot_data(self, data, units, tt):
        if units is None:
            units = "no units"

        self.out_fig = go.FigureWidget(
            data=go.Scatter3d(
                x=data[:, 0],
                y=data[:, 1],
                z=data[:, 2],
                mode="markers+lines",
                marker_color=tt,
                marker_colorscale="Viridis",
                marker_colorbar=dict(thickness=20, title="time (s)"),
                marker_size=5,
            )
        )

        self.out_fig.update_layout(
            scene=dict(
                xaxis_title=f"x ({units})",
                yaxis_title=f"y ({units})",
                zaxis_title=f"x ({units})",
            )
        )

    def update_plot(self, data, tt):
        self.out_fig.data[0].x = list(data[:, 0])
        self.out_fig.data[0].y = list(data[:, 1])
        self.out_fig.data[0].z = list(data[:, 2])
        self.out_fig.update_traces(marker_color=list(tt))


def trial_align_spatial_series(spatial_series, trials=None):
    options = [("x", 0), ("y", 1), ("z", 2)][: spatial_series.data.shape[1]]
    if spatial_series.rate is None:
        return AlignMultiTraceTimeSeriesByTrialsVariable(
            time_series=spatial_series,
            trials=trials,
            trace_controller_kwargs=dict(options=options),
        )
    else:
        return AlignMultiTraceTimeSeriesByTrialsConstant(
            time_series=spatial_series,
            trials=trials,
            trace_controller_kwargs=dict(options=options),
        )



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

from ipywidgets import widgets

from pynwb import NWBFile

from nwbwidgets.dynamictablesummary import DynamicTableSummaryWidget


class DynamicSummaryTestCase(unittest.TestCase):
    def setUp(self):
        """
        Test with unit tables
        """
        start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
        create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

        self.nwbfile = NWBFile(
            session_description="NWBFile for DynamicTableSummaryWidget",
            identifier="NWB123",
            session_start_time=start_time,
            file_create_date=create_date,
        )

        self.nwbfile.add_unit_column(
            "location", "the anatomical location of this unit"
        )
        self.nwbfile.add_unit_column(
            "sorting_quality", "the spike sorting quality of this unit"
        )
        self.nwbfile.add_unit_column(
            "quality", "the quality for the inference of this unit"
        )
        self.nwbfile.add_unit_column(
            "isi_violation", "the ISI violation ratio of this unit"
        )
        self.nwbfile.add_unit_column(
            "firing_rate", "the firing rate of this unit"
        )
        self.nwbfile.add_unit_column(
            "halfwidth", "the haldfwidth of this unit"
        )

        self.nwbfile.add_unit(
            spike_times=[2.2, 3.0, 4.5],
            obs_intervals=[[1, 10]],
            location="CA1",
            sorting_quality="good",
            quality=0.95,
            isi_violation=0.1,
            halfwidth=0.001,
            firing_rate=5.1
        )
        self.nwbfile.add_unit(
            spike_times=[2.2, 3.0, 25.0, 26.0],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA3",
            sorting_quality="ok",
            quality=0.85,
            isi_violation=0.12,
            halfwidth=0.002,
            firing_rate=2.1
        )
        self.nwbfile.add_unit(
            spike_times=[1.2, 2.3, 3.3, 4.5],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA1",
            sorting_quality="good",
            quality=0.90,
            isi_violation=0.3,
            halfwidth=0.0015,
            firing_rate=10.1
        )
        self.nwbfile.add_unit(
            spike_times=[1.4, 2.4, 3.4, 4.6],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA3",
            sorting_quality="ok",
            quality=0.92,
            isi_violation=0.21,
            halfwidth=0.0018,
            firing_rate=14.4
        )

    def test_dynamyc_summary_table_widget(self):
        widget = DynamicTableSummaryWidget(self.nwbfile.units)
        assert isinstance(widget, widgets.Widget)

        col_names_display = widget.col_names_display.keys()
        real_names = [col for col in col_names_display if "(r)" in col]
        cat_names = [col for col in col_names_display if "(c)" in col]

        # select 1 real
        widget.show_labels.value = True
        cols_to_plot = (real_names[0],)
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 1 cat
        widget.show_labels.value = True
        cols_to_plot = (cat_names[0],)
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 1 real + 1 cat
        widget.show_labels.value = True
        cols_to_plot = (real_names[0], cat_names[0])
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 2 real
        widget.show_labels.value = True
        cols_to_plot = (real_names[0], real_names[1])
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 2 real + 1 cat
        widget.show_labels.value = True
        cols_to_plot = (real_names[0], real_names[1], cat_names[0])
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 3 real
        widget.show_labels.value = True
        cols_to_plot = (real_names[0], real_names[1], real_names[2])
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # select 3 real + 1 cat
        widget.show_labels.value = True
        cols_to_plot = (real_names[0], real_names[1],
                        real_names[2], cat_names[0])
        widget.column_dropdown.value = cols_to_plot
        widget.show_labels.value = False

        # test situations that reset column dropdown

        # two categorical variables
        cols_to_plot = (cat_names[0], cat_names[1])
        widget.column_dropdown.value = cols_to_plot
        # this should trigger a reset of the dropdown
        assert len(widget.column_dropdown.value) == 0

        # two categorical + 1 real variables
        cols_to_plot = (cat_names[0], cat_names[1], real_names[0])
        widget.column_dropdown.value = cols_to_plot
        # this should trigger a reset of the dropdown
        assert len(widget.column_dropdown.value) == 0

        # 4 real variables
        cols_to_plot = (real_names[0], real_names[1],
                        real_names[2], real_names[3])
        widget.column_dropdown.value = cols_to_plot
        # this should trigger a reset of the dropdown
        assert len(widget.column_dropdown.value) == 0



from ipyvolume import TransferFunction

from nwbwidgets.utils.cmaps import (
    linear_transfer_function,
    matplotlib_transfer_function,
)


def test_linear_transfer_function():
    assert isinstance(
        linear_transfer_function("blue", reverse_opacity=True), TransferFunction
    )


def test_matplotlib_transfer_function():
    assert isinstance(matplotlib_transfer_function("PuBuGn_r"), TransferFunction)



import asyncio

import matplotlib.pyplot as plt

from ipywidgets import Output
from ipywidgets.widgets.interaction import clear_output
import plotly.graph_objects as go


def clean_axes(axes):
    """
    Removes top and right spines from axes

    Parameters
    ----------
    axes: iterable

    """
    for ax in axes:
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)


def unpack_controls(controls, process_controls=lambda x: x):
    control_states = {}
    for k, v in controls.items():
        # if the value is a dict, add those individually
        if isinstance(v.value, dict):
            control_states.update(v.value)
        else:
            control_states[k] = v.value
    kwargs = process_controls(control_states)
    return kwargs


def interactive_output(f, controls, process_controls=lambda x: x, fixed=None):
    """Connect widget controls to a function.

    This function does not generate a user interface for the widgets (unlike `interact`).
    This enables customisation of the widget user interface layout.
    The user interface layout must be defined and displayed manually.
    """

    if fixed is None:
        fixed = dict()

    out = Output()

    def observer(change):
        with out:
            clear_output(wait=True)
            plot = f(**fixed, **unpack_controls(controls, process_controls))
            plt.show()

    for k, w in controls.items():
        w.observe(observer, "value")
    observer(None)
    return out

def set_plotly_callbacks(f, controls, process_controls=lambda x: x):
    fig = go.FigureWidget()

    def observer(change):
        return f(fig=fig, **unpack_controls(controls, process_controls))
    for k, w in controls.items():
        w.observe(observer, "value")
    return observer(None)

class Timer:
    def __init__(self, timeout, callback):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())

    async def _job(self):
        await asyncio.sleep(self._timeout)
        self._callback()

    def cancel(self):
        self._task.cancel()


def debounce(wait):
    """Decorator that will postpone a function's
    execution until after `wait` seconds
    have elapsed since the last time it was invoked."""

    def decorator(fn):
        timer = None

        def debounced(*args, **kwargs):
            nonlocal timer

            def call_it():
                fn(*args, **kwargs)

            if timer is not None:
                timer.cancel()
            timer = Timer(wait, call_it)

        return debounced

    return decorator



import numpy as np
from scipy.signal import stft
import matplotlib.pyplot as plt

import plotly.graph_objects as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from ipywidgets import widgets, ValueWidget

from pynwb import TimeSeries
from pynwb.base import DynamicTable
from pynwb.ecephys import SpikeEventSeries, ElectricalSeries

from .base import fig2widget, lazy_tabs, render_dataframe
from .timeseries import BaseGroupedTraceWidget
from .brains import HumanElectrodesPlotlyWidget


def show_spectrogram(nwbobj: TimeSeries, channel=0, **kwargs):
    fig, ax = plt.subplots()
    f, t, Zxx = stft(nwbobj.data[:, channel], nwbobj.rate, nperseg=2 * 17)
    ax.imshow(
        np.log(np.abs(Zxx)),
        aspect="auto",
        extent=[0, max(t), 0, max(f)],
        origin="lower",
    )
    ax.set_ylim(0, max(f))
    ax.set_xlabel("time (s)")
    ax.set_ylabel("frequency (Hz)")
    fig.show()


class ElectrodeGroupsWidget(ValueWidget, widgets.HBox):
    def __init__(self, nwbobj: DynamicTable, **kwargs):
        super().__init__()
        group_names = nwbobj.group_name[:]
        ugroups, group_pos, counts = np.unique(
            group_names, return_inverse=True, return_counts=True
        )
        elec_pos = np.hstack(np.arange(count).tolist() for count in counts)

        hovertext = []
        df = nwbobj.to_dataframe()
        for i, row in df.iterrows():
            hovertext.append("")
            for key, val in list(row.to_dict().items()):
                if key == "group":
                    continue
                hovertext[-1] += "{}: {}<br>".format(key, val)

        self.fig = go.FigureWidget()
        self.fig.add_trace(
            go.Scatter(
                x=elec_pos,
                y=nwbobj.group_name[:],
                mode="markers",
                marker=dict(
                    color=np.array(DEFAULT_PLOTLY_COLORS)[
                        group_pos % len(DEFAULT_PLOTLY_COLORS)
                    ],
                    size=15,
                ),
                hovertext=hovertext,
                hoverinfo="text",
            )
        )

        self.fig.update_layout(
            width=400,
            height=700,
            xaxis_title="group electrode number",
            showlegend=False,
            margin=dict(t=10),
        )

        self.value = list(range(len(group_names)))

        def selection_fn(trace, points, selector):
            self.value = points.point_inds

        self.fig.data[0].on_selection(selection_fn)

        self.children = [self.fig]


def show_electrodes(electrodes_table):
    in_dict = dict(table=render_dataframe)
    if np.isnan(electrodes_table.x[0]):  # position is not defined
        in_dict.update(electrode_groups=ElectrodeGroupsWidget)
    else:
        subject = electrodes_table.get_ancestor("NWBFile").subject
        if subject is not None:
            species = subject.species
            if species in ("mouse", "Mus musculus"):
                in_dict.update(CCF=show_ccf)
            elif species in ("human", "Homo sapiens"):
                in_dict.update(render=HumanElectrodesPlotlyWidget)

    return lazy_tabs(in_dict, electrodes_table)


def show_ccf(electrodes_table=None, **kwargs):
    from ccfwidget import CCFWidget

    input_kwargs = {}
    if electrodes_table is not None:
        df = electrodes_table.to_dataframe()
        markers = [
            idf[["x", "y", "z"]].to_numpy() for _, idf in df.groupby("group_name")
        ]
        input_kwargs.update(markers=markers)

    input_kwargs.update(kwargs)
    return CCFWidget(**input_kwargs)


def show_spike_event_series(ses: SpikeEventSeries, **kwargs):
    def control_plot(spk_ind):
        fig, ax = plt.subplots(figsize=(9, 5))
        data = ses.data[spk_ind]
        if nChannels > 1:
            for ch in range(nChannels):
                ax.plot(data[:, ch], color="#d9d9d9")
        else:
            ax.plot(data[:], color="#d9d9d9")
        ax.plot(np.mean(data, axis=1), color="k")
        ax.set_xlabel("Time")
        ax.set_ylabel("Amplitude")
        fig.show()
        return fig2widget(fig)

    if len(ses.data.shape) == 3:
        nChannels = ses.data.shape[2]
    else:
        nChannels = ses.data.shape[1]
    nSpikes = ses.data.shape[0]

    # Controls
    field_lay = widgets.Layout(
        max_height="40px", max_width="100px", min_height="30px", min_width="70px"
    )
    spk_ind = widgets.BoundedIntText(value=0, min=0, max=nSpikes - 1, layout=field_lay)
    controls = {"spk_ind": spk_ind}
    out_fig = widgets.interactive_output(control_plot, controls)

    # Assemble layout box
    lbl_spk = widgets.Label("Spike ID:", layout=field_lay)
    lbl_nspks0 = widgets.Label("N° spikes:", layout=field_lay)
    lbl_nspks1 = widgets.Label(str(nSpikes), layout=field_lay)
    lbl_nch0 = widgets.Label("N° channels:", layout=field_lay)
    lbl_nch1 = widgets.Label(str(nChannels), layout=field_lay)
    hbox0 = widgets.HBox(children=[lbl_spk, spk_ind])
    vbox0 = widgets.VBox(
        children=[
            widgets.HBox(children=[lbl_nspks0, lbl_nspks1]),
            widgets.HBox(children=[lbl_nch0, lbl_nch1]),
            hbox0,
        ]
    )
    hbox1 = widgets.HBox(children=[vbox0, out_fig])

    return hbox1


class ElectricalSeriesWidget(BaseGroupedTraceWidget):
    def __init__(
        self,
        electrical_series: ElectricalSeries,
        neurodata_vis_spec=None,
        foreign_time_window_controller=None,
        foreign_group_and_sort_controller=None,
        dynamic_table_region_name="electrodes",
        **kwargs
    ):
        if foreign_group_and_sort_controller is not None:
            table = None
        else:
            table = dynamic_table_region_name
        super().__init__(
            electrical_series,
            table,
            foreign_time_window_controller=foreign_time_window_controller,
            foreign_group_and_sort_controller=foreign_group_and_sort_controller,
            **kwargs
        )



import numpy as np
import matplotlib.pyplot as plt

from ipywidgets import widgets

from ndx_spectrum import Spectrum


def show_spectrum(node: Spectrum, **kwargs) -> widgets.Widget:
    check_spectrum(node)
    data = node.power if "power" in node.fields else node.phase
    if len(data.shape) == 2:
        no_channels = data.shape[1]
    else:
        no_channels = 1
    freqs_all = np.asarray(node.frequencies)
    channel_slider = widgets.IntRangeSlider(
        min=1,
        max=no_channels,
        step=1,
        description="Channel Range:",
        disabled=False,
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format="d",
    )
    freq_slider = widgets.IntRangeSlider(
        min=0,
        max=freqs_all[-1],
        step=1,
        description="Frequency Range:",
        disabled=False,
        continuous_update=False,
        orientation="horizontal",
        readout=True,
        readout_format="d",
    )
    out = widgets.Output()
    with out:
        widgets.interact(
            lambda channel_range, frequency_range: plot_spectrum_figure(
                node, channel_range, frequency_range
            ),
            channel_range=channel_slider,
            frequency_range=freq_slider,
        )
    return out


def plot_spectrum_figure(spectrum, channel_nos, frequency_nos):
    """
    Plot power vs frequencies and/or phase vs frequencies.
    Parameters
    ----------
    spectrum: Spectrum
    channel_nos: tuple
        Input from the channel range slider widget: (channel_no start, channel_no end)
    frequency_nos: tuple
        Input from frequency range slider widget: (freq start, freq end)
    """
    check_spectrum(spectrum)
    all_freqs = np.asarray(spectrum.frequencies)
    start_id = (np.abs(frequency_nos[0] - all_freqs)).argmin()
    end_id = (np.abs(frequency_nos[1] - all_freqs)).argmin()
    if channel_nos[0] != channel_nos[1]:
        range_ = range(channel_nos[0] - 1, channel_nos[1] - 1)
    else:
        range_ = channel_nos[0] - 1
    if "power" in spectrum.fields:
        power_data = np.asarray(spectrum.power)
        if len(power_data.shape) == 1:
            power_data = power_data[:, np.newaxis]
    if "phase" in spectrum.fields:
        phase_data = np.asarray(spectrum.phase)
        if len(phase_data.shape) == 1:
            phase_data = phase_data[:, np.newaxis]
    # make plots:
    if "power" in spectrum.fields and "phase" in spectrum.fields:
        fig, axs = plt.subplots(2, 1, sharex=True)
        axs[0].semilogy(all_freqs[start_id:end_id], power_data[start_id:end_id, range_])
        axs[0].set_ylabel("Power")
        axs[1].plot(all_freqs[start_id:end_id], phase_data[start_id:end_id, range_])
        axs[1].set_ylabel("phase")
    elif "power" in spectrum.fields:
        fig, ax = plt.subplots()
        ax.set_xlabel("frequency")
        ax.semilogy(all_freqs[start_id:end_id], power_data[start_id:end_id, range_])
        ax.set_ylabel("Power")
    elif "phase" in spectrum.fields:
        fig, ax = plt.subplots()
        ax.plot(all_freqs[start_id:end_id], phase_data[start_id:end_id, range_])
        ax.set_ylabel("phase")
        ax.set_xlabel("frequency")


def check_spectrum(node):
    assert isinstance(node, Spectrum), "datatype not of type Spectrum"
    assert "frequencies" in node.fields, "frequencies not found in Spectrum"
    assert (
        "power" in node.fields or "phase" in node.fields
    ), "neither of power/phase found in Spectrum"
    if "power" in node.fields and "phase" in node.fields:
        assert (
            node.power.shape == node.phase.shape
        ), "power and phase arrays should be of same dims"



from pathlib import Path, PureWindowsPath

import numpy as np

import plotly.graph_objects as go
from ipywidgets import widgets, Layout

from tifffile import imread

from nwbwidgets.image import ImageSeriesWidget
from nwbwidgets.timeseries import SingleTracePlotlyWidget
from nwbwidgets.controllers import StartAndDurationController
from nwbwidgets.utils.timeseries import get_timeseries_maxt, get_timeseries_mint


class AllenDashboard(widgets.VBox):
    def __init__(self, nwb):
        super().__init__()
        self.nwb = nwb
        self.show_spikes = False

        self.btn_spike_times = widgets.Button(
            description="Show spike times", button_style=""
        )
        self.btn_spike_times.on_click(self.spikes_viewer)

        # Start time and duration controller
        self.tmin = get_timeseries_mint(
            nwb.processing["ophys"]
            .data_interfaces["fluorescence"]
            .roi_response_series["roi_response_series"]
        )
        self.tmax = get_timeseries_maxt(
            nwb.processing["ophys"]
            .data_interfaces["fluorescence"]
            .roi_response_series["roi_response_series"]
        )
        self.time_window_controller = StartAndDurationController(
            tmin=self.tmin,
            tmax=self.tmax,
            start=0,
            duration=5,
        )

        # Electrophys single trace
        self.electrical = SingleTracePlotlyWidget(
            timeseries=nwb.processing["ecephys"].data_interfaces[
                "filtered_membrane_voltage"
            ],
            foreign_time_window_controller=self.time_window_controller,
        )
        self.electrical.out_fig.update_layout(
            title=None,
            showlegend=False,
            xaxis_title=None,
            yaxis_title="Volts",
            width=840,
            height=230,
            margin=dict(l=60, r=200, t=8, b=20),
            # yaxis={"position": 0, "anchor": "free"},
            yaxis={
                "range": [
                    min(self.electrical.out_fig.data[0].y),
                    max(self.electrical.out_fig.data[0].y),
                ],
                "autorange": False,
            },
            xaxis={
                "showticklabels": False,
                "ticks": "",
                "range": [
                    min(self.electrical.out_fig.data[0].x),
                    max(self.electrical.out_fig.data[0].x),
                ],
                "autorange": False,
            },
        )
        # Fluorescence single trace
        self.fluorescence = SingleTracePlotlyWidget(
            timeseries=nwb.processing["ophys"]
            .data_interfaces["fluorescence"]
            .roi_response_series["roi_response_series"],
            foreign_time_window_controller=self.time_window_controller,
        )
        self.fluorescence.out_fig.update_layout(
            title=None,
            showlegend=False,
            width=840,
            height=230,
            margin=dict(l=60, r=200, t=8, b=20),
            yaxis_title="dF/F",
            yaxis={
                "range": [
                    min(self.fluorescence.out_fig.data[0].y),
                    max(self.fluorescence.out_fig.data[0].y),
                ],
                "autorange": False,
            },
            xaxis={
                "range": [
                    min(self.fluorescence.out_fig.data[0].x),
                    max(self.fluorescence.out_fig.data[0].x),
                ],
                "autorange": False,
                "constrain": "domain",
                "anchor": "free",
            },
        )
        # Two photon imaging
        self.photon_series = ImageSeriesWidget(
            imageseries=nwb.acquisition["raw_ophys"],
            foreign_time_window_controller=self.time_window_controller,
        )
        self.photon_series.out_fig.update_layout(
            showlegend=False,
            margin=dict(l=30, r=5, t=65, b=65),
        )

        # Frame controller
        self.frame_controller = widgets.FloatSlider(
            value=0,
            step=1 / self.nwb.acquisition["raw_ophys"].rate,
            min=self.time_window_controller.value[0],
            max=self.time_window_controller.value[1],
            description="Frame: ",
            style={"description_width": "55px"},
            continuous_update=False,
            readout=False,
            orientation="horizontal",
            layout=Layout(width="645px"),
        )

        # Add line traces marking Image frame point
        self.frame_point = go.Scatter(x=[0, 0], y=[-1000, 1000])
        self.electrical.out_fig.add_trace(self.frame_point)
        self.fluorescence.out_fig.add_trace(self.frame_point)

        # Updates frame point
        self.frame_controller.observe(self.update_frame_point)

        # Updates list of valid spike times at each change in time range
        self.time_window_controller.observe(self.updated_time_range)

        # Layout
        hbox_header = widgets.HBox([self.btn_spike_times, self.time_window_controller])
        vbox_widgets = widgets.VBox([self.electrical, self.fluorescence])
        hbox_widgets = widgets.HBox([vbox_widgets, self.photon_series])

        self.children = [hbox_header, self.frame_controller, hbox_widgets]

        self.update_spike_traces()

    def update_frame_point(self, change):
        """Updates Image frame and frame point relative position on temporal traces"""
        if isinstance(change["new"], float):
            self.electrical.out_fig.data[1].x = [change["new"], change["new"]]
            self.fluorescence.out_fig.data[1].x = [change["new"], change["new"]]

            frame_number = int(change["new"] * self.nwb.acquisition["raw_ophys"].rate)
            file_path = self.nwb.acquisition["raw_ophys"].external_file[0]
            if "\\" in file_path:
                win_path = PureWindowsPath(file_path)
                path_ext_file = Path(win_path)
            else:
                path_ext_file = Path(file_path)
            image = imread(path_ext_file, key=frame_number)
            self.photon_series.out_fig.data[0].z = image

    def updated_time_range(self, change=None):
        """Operations to run whenever time range gets updated"""
        self.update_spike_traces()
        self.show_spikes = False

        # Update frame slider
        if self.time_window_controller.value[1] < self.frame_controller.min:
            self.frame_controller.min = self.time_window_controller.value[0]
            self.frame_controller.max = self.time_window_controller.value[1]
        else:
            self.frame_controller.max = self.time_window_controller.value[1]
            self.frame_controller.min = self.time_window_controller.value[0]
        xpoint = round(np.mean(self.time_window_controller.value))
        self.frame_controller.value = xpoint
        self.electrical.out_fig.data[1].x = [xpoint, xpoint]
        self.fluorescence.out_fig.data[1].x = [xpoint, xpoint]

        # Reset spike times view
        self.btn_spike_times.description = "Show spike times"
        self.fluorescence.out_fig.data = [
            self.fluorescence.out_fig.data[0],
            self.fluorescence.out_fig.data[1],
        ]
        self.electrical.out_fig.data = [
            self.electrical.out_fig.data[0],
            self.electrical.out_fig.data[1],
        ]

    def spikes_viewer(self, b=None):
        self.show_spikes = not self.show_spikes
        if self.show_spikes:
            self.btn_spike_times.description = "Hide spike times"
            for spike_trace in self.spike_traces:
                self.fluorescence.out_fig.add_trace(spike_trace)
        else:
            self.btn_spike_times.description = "Show spike times"
            self.fluorescence.out_fig.data = [
                self.fluorescence.out_fig.data[0],
                self.fluorescence.out_fig.data[1],
            ]

    def update_spike_traces(self):
        """Updates list of go.Scatter objects at spike times"""
        self.spike_traces = []
        t_start = self.time_window_controller.value[0]
        t_end = self.time_window_controller.value[1]
        all_spikes = self.nwb.units["spike_times"][0]
        mask = (all_spikes > t_start) & (all_spikes < t_end)
        selected_spikes = all_spikes[mask]
        # Makes a go.Scatter object for each spike in chosen interval
        for spkt in selected_spikes:
            self.spike_traces.append(
                go.Scatter(
                    x=[spkt, spkt],
                    y=[-1000, 1000],
                    line={"color": "gray", "width": 0.5},
                )
            )



from bisect import bisect_right, bisect_left

import numpy as np

from pynwb.misc import Units


def get_spike_times(units: Units, index, in_interval):
    """Use bisect methods to efficiently retrieve spikes from a given unit in a given interval

    Parameters
    ----------
    units: pynwb.misc.Units
    index: int
    in_interval: start and stop times

    Returns
    -------

    """
    st = units["spike_times"]
    unit_start = 0 if index == 0 else st.data[index - 1]
    unit_stop = st.data[index]
    start_time, stop_time = in_interval

    ind_start = bisect_left(st.target, start_time, unit_start, unit_stop)
    ind_stop = bisect_right(st.target, stop_time, ind_start, unit_stop)

    return np.asarray(st.target[ind_start:ind_stop])


def get_min_spike_time(units: Units):
    """Efficiently retrieve the first spike time across all units

    Parameters
    ----------
    units: pynwb.misc.Units

    Returns
    -------

    """
    st = units["spike_times"]
    inds = [0] + list(st.data[:-1])
    first_spikes = [st.target.data[i] for i in inds]
    return np.min(first_spikes)


def get_max_spike_time(units: Units):
    """Efficiently retrieve the last spike time across all units

    Parameters
    ----------
    units: pynwb.misc.Units

    Returns
    -------

    """
    st = units["spike_times"]
    inds = [x - 1 for x in st.data[:]]
    last_spikes = [st.target.data[i] for i in inds]
    return np.max(last_spikes)


def align_by_times(units: Units, index, starts, stops):
    """
    Args:
        units: pynwb.misc.Units
        index: int
        starts: array-like
        stops: array-like
    Returns:
        np.array
    """

    st = units["spike_times"]
    unit_spike_data = st[index]

    istarts = np.searchsorted(unit_spike_data, starts)
    istops = np.searchsorted(unit_spike_data, stops)
    for start, istart, istop in zip(starts, istarts, istops):
        yield unit_spike_data[istart:istop] - start


def align_by_trials(
    units: Units,
    index,
    start_label="start_time",
    stop_label=None,
    start=-0.5,
    end=1.0,
):
    """
    Args:
        units
        start_label: str
            default: 'start_time'
        stop_label: str
            default: None (just align to start_time)
        start: float
            Start time for calculation before or after (negative or positive) the reference point (aligned to).
        end: float
            End time for calculation before or after (negative or positive) the reference point (aligned to).
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    trials = units.get_ancestor("NWBFile").trials
    return align_by_time_intervals(
        units, index, trials, start_label, stop_label, start, end
    )


def align_by_time_intervals(
    units: Units,
    index,
    intervals,
    start_label="start_time",
    stop_label="stop_time",
    start=0.0,
    end=0.0,
    rows_select=(),
    progress_bar=None,
):
    """
    Args:
        units: time-aware neurodata_type
        index: int
        intervals: pynwb.epoch.TimeIntervals
        start_label: str
            default: 'start_time'
        stop_label: str
            default: 'stop_time'
        start: float
            Start time for calculation before or after (negative or positive) the reference point (aligned to).
        end: float
            End time for calculation before or after (negative or positive) the reference point (aligned to).
        rows_select: array_like, optional
            sub-selects specific rows
        progress_bar: FloatProgress, optional
            Proved a progress bar object to have this method automatically update the progress bar
    Returns:
        np.array(shape=(n_trials, n_time, ...))
    """
    if stop_label is None:
        stop_label = start_label
    starts = np.squeeze(np.array(intervals[start_label][:])[rows_select] + start)
    stops = np.squeeze(np.array(intervals[stop_label][:])[rows_select] + end)
    if progress_bar is not None:
        progress_bar.value = 0
        progress_bar.description = "reading spike data"

    out = []
    for i, x in enumerate(align_by_times(units, index, starts, stops)):
        out.append(x + start)
        if progress_bar is not None:
            progress_bar.value = i / len(units)

    return out


def get_unobserved_intervals(units, time_window, units_select=()):

    if "obs_intervals" not in units:
        return []

    # add observation intervals
    unobserved_intervals_list = []
    for i_unit in units_select:
        intervals = units["obs_intervals"][i_unit]  # TODO: use bisect here
        intervals = np.array(intervals, dtype="object")
        these_obs_intervals = intervals[
            (intervals[:, 1] > time_window[0]) & (intervals[:, 0] < time_window[1])
        ]
        unobs_intervals = np.c_[these_obs_intervals[:-1, 1], these_obs_intervals[1:, 0]]

        if len(these_obs_intervals):
            # handle unobserved interval on lower bound of window
            if these_obs_intervals[0, 0] > time_window[0]:
                unobs_intervals = np.vstack(
                    ([time_window[0], these_obs_intervals[0, 0]], unobs_intervals)
                )

            # handle unobserved interval on lower bound of window
            if these_obs_intervals[-1, 1] < time_window[1]:
                unobs_intervals = np.vstack(
                    (unobs_intervals, [these_obs_intervals[-1, 1], time_window[1]])
                )
        else:
            unobs_intervals = [time_window]

        unobserved_intervals_list.append(unobs_intervals)

    return unobserved_intervals_list



import unittest
import pytest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ipywidgets import widgets

from pynwb import NWBFile, ProcessingModule, TimeSeries
from pynwb.behavior import Position, SpatialSeries
from pynwb.core import DynamicTable
from pynwb.file import Subject

from nwbwidgets.base import (
    show_neurodata_base,
    processing_module,
    nwb2widget,
    show_text_fields,
    fig2widget,
    vis2widget,
    show_fields,
    df2accordion,
    lazy_show_over_data,
)
from nwbwidgets.view import default_neurodata_vis_spec, show_dynamic_table


def test_show_neurodata_base():
    start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
    create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="demonstrate NWBFile basics",
        identifier="NWB123",
        session_start_time=start_time,
        file_create_date=create_date,
        related_publications="https://doi.org/10.1088/1741-2552/aaa904",
        experimenter="Dr. Pack",
    )

    assert isinstance(
        show_neurodata_base(nwbfile, default_neurodata_vis_spec), widgets.Widget
    )


def test_show_text_fields():
    data = np.random.rand(160, 3)
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
    )
    assert isinstance(show_text_fields(ts), widgets.Widget)


class ProcessingModuleTestCase(unittest.TestCase):
    def setUp(self):
        spatial_series = SpatialSeries(
            name="position",
            data=np.linspace(0, 1, 20),
            rate=50.0,
            reference_frame="starting gate",
        )
        self.position = Position(spatial_series=spatial_series)

    def test_processing_module(self):
        start_time = datetime(2020, 1, 29, 11, tzinfo=tzlocal())
        nwbfile = NWBFile(
            session_description="Test Session",
            identifier="NWBPM",
            session_start_time=start_time,
        )

        behavior_module = ProcessingModule(
            name="behavior", description="preprocessed behavioral data"
        )
        nwbfile.add_processing_module(behavior_module)

        nwbfile.processing["behavior"].add(self.position)

        processing_module(nwbfile.processing["behavior"], default_neurodata_vis_spec)

    def test_nwb2widget(self):
        nwb2widget(self.position, default_neurodata_vis_spec)


def test_fig2widget():
    data = np.random.rand(160, 3)
    fig = plt.figure(figsize=(10, 5))
    plt.plot(data)

    assert isinstance(fig2widget(fig), widgets.Widget)


class TestVis2widget:
    def test_vis2widget_input_widget(self):
        wg = widgets.IntSlider(
            value=7,
            min=0,
            max=10,
            step=1,
            description="Test:",
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="d",
        )

        assert isinstance(vis2widget(wg), widgets.Widget)

    def test_vis2widget_input_figure(self):
        data = np.random.rand(160, 3)

        fig = plt.figure(figsize=(10, 5))
        plt.plot(data)

        assert isinstance(vis2widget(fig), widgets.Widget)

    def test_vis2widget_input_other(self):
        data = np.random.rand(160, 3)
        with pytest.raises(ValueError, match="unsupported vis type"):
            vis2widget(data)


def test_show_subject():
    node = Subject(age="8", sex="m", species="macaque")
    show_fields(node)


def test_show_dynamic_table():
    d = {"col1": [1, 2], "col2": [3, 4]}
    dt = DynamicTable.from_dataframe(
        df=pd.DataFrame(data=d), name="Test Dtable", table_description="no description"
    )
    show_dynamic_table(dt)


def test_df2accordion():
    df = pd.DataFrame(
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), columns=["a", "b", "c"]
    )

    def func_fig(data):
        fig = plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig

    df2accordion(df=df, by="a", func=func_fig)


def test_df2accordion_single():
    df = pd.DataFrame(np.array([1]), columns=["a"])

    def func_fig(data):
        fig = plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig

    df2accordion(df=df, by="a", func=func_fig)


def test_lazy_show_over_data():
    list_ = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    def func_fig(data):
        fig = plt.figure(figsize=(10, 5))
        plt.plot(data)
        return fig

    assert isinstance(lazy_show_over_data(list_=list_, func_=func_fig), widgets.Widget)



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np
import matplotlib.pyplot as plt

from ipywidgets import widgets

from pynwb import NWBFile
from pynwb.misc import DecompositionSeries, AnnotationSeries

from nwbwidgets.misc import (
    show_psth_raster,
    PSTHWidget,
    show_decomposition_traces,
    show_decomposition_series,
    RasterWidget,
    show_session_raster,
    show_annotations,
    RasterGridWidget,
    raster_grid,
)


def test_show_psth():
    data = np.random.random([6, 50])
    assert isinstance(show_psth_raster(data=data, start=0, end=1), plt.Subplot)


def test_show_annotations():
    timestamps = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    annotations = AnnotationSeries(name="test_annotations", timestamps=timestamps)
    show_annotations(annotations)


class ShowPSTHTestCase(unittest.TestCase):
    def setUp(self):
        """
        Trials must exist.
        """
        start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
        create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

        self.nwbfile = NWBFile(
            session_description="NWBFile for PSTH",
            identifier="NWB123",
            session_start_time=start_time,
            file_create_date=create_date,
        )

        self.nwbfile.add_unit_column("location", "the anatomical location of this unit")
        self.nwbfile.add_unit_column(
            "quality", "the quality for the inference of this unit"
        )

        self.nwbfile.add_unit(
            spike_times=[2.2, 3.0, 4.5],
            obs_intervals=[[1, 10]],
            location="CA1",
            quality=0.95,
        )
        self.nwbfile.add_unit(
            spike_times=[2.2, 3.0, 25.0, 26.0],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA3",
            quality=0.85,
        )
        self.nwbfile.add_unit(
            spike_times=[1.2, 2.3, 3.3, 4.5],
            obs_intervals=[[1, 10], [20, 30]],
            location="CA1",
            quality=0.90,
        )

        self.nwbfile.add_trial_column(
            name="stim", description="the visual stimuli during the trial"
        )

        self.nwbfile.add_trial(start_time=0.0, stop_time=1.0, stim="person")
        self.nwbfile.add_trial(start_time=0.1, stop_time=2.0, stim="person")
        self.nwbfile.add_trial(start_time=3.0, stop_time=4.0, stim="ocean")
        self.nwbfile.add_trial(start_time=4.0, stop_time=5.0, stim="ocean")
        self.nwbfile.add_trial(start_time=5.0, stop_time=6.0, stim="desert")
        self.nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim="desert")

    def test_psth_widget(self):
        widget = PSTHWidget(self.nwbfile.units)
        assert isinstance(widget, widgets.Widget)

        widget.psth_type_radio = "gaussian"
        widget.trial_event_controller.value = ("start_time", "stop_time")
        widget.unit_controller.value = 1
        widget.gas.group_dd.value = "stim"
        widget.gas.group_dd.value = None

    def test_multipsth_widget(self):
        psth_widget = PSTHWidget(self.nwbfile.units)
        assert isinstance(psth_widget, widgets.Widget)
        start_labels = ('start_time', 'stop_time')
        fig = psth_widget.update(index=0, start_labels=start_labels)
        assert len(fig.axes) == 2 * len(start_labels)

    def test_raster_widget(self):
        assert isinstance(RasterWidget(self.nwbfile.units), widgets.Widget)

    def test_show_session_raster(self):
        assert isinstance(show_session_raster(self.nwbfile.units), plt.Axes)

    def test_raster_grid_widget(self):
        assert isinstance(RasterGridWidget(self.nwbfile.units), widgets.Widget)

    def test_raster_grid(self):
        trials = self.nwbfile.units.get_ancestor("NWBFile").trials
        assert isinstance(
            raster_grid(
                self.nwbfile.units,
                time_intervals=trials,
                index=0,
                start=-0.5,
                end=20.0,
            ),
            plt.Figure,
        )


class ShowDecompositionTestCase(unittest.TestCase):
    def setUp(self):
        data = np.random.rand(160, 2, 3)

        self.ds = DecompositionSeries(
            name="Test Decomposition", data=data, metric="amplitude", rate=1.0
        )

    def test_show_decomposition_traces(self):
        assert isinstance(show_decomposition_traces(self.ds), widgets.Widget)

    def test_show_decomposition_series(self):
        assert isinstance(show_decomposition_series(self.ds), widgets.Widget)



"""Pytest configuration file."""

import matplotlib.pyplot as plt

def pytest_configure(config):
    """Set configurations to be run before all testing."""

    # Turn on interactive mode for matplotlib
    #   In practice, this defaults `block=False` for plt.show()
    #   This means that plots won't show and wait for input before proceeding
    #   To run tests while showing plots, comment this out, or replace with `plt.ioff`
    plt.ion()



import os
from setuptools import setup
from pathlib import Path

root = Path(__file__).parent

d = {}
exec(open("nwbwidgets/version.py").read(), None, d)
version = d["version"]

with open("README.md") as f:
    long_description = f.read()

with open(os.path.join(root, "requirements.txt")) as f:
    requirements = f.readlines()

setup(
    author="Ben Dichter",
    author_email="ben.dichter@catalystneuro.com",
    version=version,
    classifiers=[
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Framework :: Jupyter",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="This is nwbwidgets, widgets for viewing the contents of a "
    "NWB-file in Jupyter Notebooks using ipywidgets.",
    install_requires=requirements,
    license="BSD",
    keywords=["jupyter", "hdf5", "notebook", "nwb"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="nwbwidgets",
    packages=[
        "nwbwidgets",
        "nwbwidgets/utils",
        "nwbwidgets/analysis",
        "nwbwidgets/controllers",
    ],
    python_requires=">=3.7",
    setup_requires=["setuptools>=38.6.0", "setuptools_scm"],
    url="https://github.com/NeurodataWithoutBorders/nwb-jupyter-widgets",
)



from typing import Iterable

import numpy as np

from hdmf.common import DynamicTable

from ipywidgets import widgets, Layout, ValueWidget

from ..utils.dynamictable import group_and_sort, infer_categorical_columns
from ..utils.pynwb import robust_unique
from . import RangeController


class AbstractGroupAndSortController(widgets.VBox, ValueWidget):
    """
    Defines the abstract type for GroupAndSortController objects. These classes take in a DynamicTable objects
    and broadcast a `value` of the form
    dict(
        order=array-like(uint),
        group_inds=array-like(uint) | None,
        labels=array-like(str) | None
    )
    """

    def __init__(self, dynamic_table: DynamicTable, keep_rows=None):
        super().__init__()
        self.keep_rows = (
            keep_rows if keep_rows is not None else np.arange(len(dynamic_table))
        )
        self.dynamic_table = dynamic_table
        self.nitems = len(self.keep_rows)
        self.column_values = None
        self.group_by = None
        self.selected_column_values = None
        self.limit = None
        self.desc = False
        self.order_by = None
        self.order_vals = None
        self.window = None


class GroupAndSortController(AbstractGroupAndSortController):
    def __init__(
        self,
        dynamic_table: DynamicTable,
        group_by=None,
        window=None,
        keep_rows=None,
        control_order=True,
        control_limit=True,
        groups=None,
    ):
        """

        Parameters
        ----------
        dynamic_table: DynamicTable
            the table wrt which the grouping is performed
        group_by: str
            the column name from the dynamic table for which the grouping is performed
        window: None or bool,
        keep_rows: Iterable
            rows of dynamic table to consider in the grouping op
        control_limit: bool
            whether to control the limit of the displayed rows
        control_order: bool
            whether to control the order of the displayed rows based on other column
        groups: dict
            dict(column_name=column_values) to work with specific columns only
        """
        super().__init__(dynamic_table, keep_rows)

        self.control_order = control_order
        self.control_limit = control_limit

        self.categorical_columns = self.get_groups() if groups is None else groups
        self.limit_bit = None
        self.limit_cb = None
        self.order_dd = None
        self.ascending_dd = None
        self.group_sm = None
        self.group_dd = None

        if len(self.categorical_columns) > 0:
            if control_limit:
                self.limit_cb = widgets.Checkbox(
                    description="limit",
                    style={"description_width": "initial"},
                    disabled=True,
                    indent=False,
                    layout=Layout(max_width="70px"),
                )
                self.limit_cb.observe(self.limit_cb_observer)
                self.limit_bit = widgets.BoundedIntText(
                    value=50,
                    min=0,
                    max=99999,
                    disabled=True,
                    layout=Layout(max_width="70px"),
                )
                self.limit_bit.observe(self.limit_bit_observer)

            self.order_dd = widgets.Dropdown(
                options=[None] + list(self.categorical_columns),
                description="order by",
                layout=Layout(max_width="120px"),
                style={"description_width": "initial"},
                disabled=not len(self.categorical_columns),
            )
            self.order_dd.observe(self.order_dd_observer)

            self.ascending_dd = widgets.Dropdown(
                options=["ASC", "DESC"], disabled=True, layout=Layout(max_width="70px")
            )
            self.ascending_dd.layout.height = "0px"
            self.ascending_dd.layout.visibility = "hidden"
            self.ascending_dd.observe(self.ascending_dd_observer)

            self.group_sm = widgets.SelectMultiple(
                layout=Layout(width="0px"),
                disabled=True,
                rows=1
            )
            self.group_sm.layout.visibility = "hidden"
            self.group_sm.observe(self.group_sm_observer)

            if group_by is None and len(self.categorical_columns) > 0:
                self.group_dd = widgets.Dropdown(
                    options=[None] + list(self.categorical_columns),
                    description="group by",
                    style={"description_width": "initial"},
                    layout=Layout(width="90%"),
                    disabled=not len(self.categorical_columns),
                )
                self.group_dd.observe(self.group_dd_observer)
            else:
                self.group_dd = None
                self.set_group_by(group_by)

        if window is None:
            range_controller_max = min(30, self.nitems)
            dt_desc_map = {
                "DynamicTable": "traces",
                "TimeIntervals": "trials",
                "Units": "units",
                "PlaneSegmentation": "image_planes",
            }
            desc = (
                dt_desc_map[dynamic_table.neurodata_type]
                if dynamic_table is not None
                else "traces"
            )
            self.range_controller = RangeController(
                0,
                self.nitems,
                start_value=(0, range_controller_max),
                dtype="int",
                description=desc,
                orientation="vertical",
            )
            self.range_controller.layout = Layout(min_width="75px")
            self.range_controller.observe(self.range_controller_observer)
            self.window = self.range_controller.value
        elif window is False:
            self.window = (0, self.nitems)
            self.range_controller = widgets.HTML("")

        self.children = self.get_children()
        #self.layout = Layout(overflow_x='auto')
        self.update_value()

    def get_children(self):
        children = []

        if self.group_dd:
            children.append(self.group_dd)
        if self.group_sm is not None:
            children.append(
                widgets.HBox(children=(self.group_sm, self.range_controller))
            )
        else:
            children.append(self.range_controller)
        if len(self.categorical_columns) > 0:
            if self.control_limit:
                children.append(
                    widgets.VBox(
                        children=(self.limit_cb, self.limit_bit),
                        layout=Layout(max_width="90%"),
                    )
                )

            if self.control_order:
                children.append(
                    widgets.VBox(
                        children=(self.order_dd, self.ascending_dd),
                        #layout=Layout(max_width="90%"),
                    )
                )

        return children

    def set_group_by(self, group_by):
        self.group_by = group_by
        self.column_values = self.get_column_values(column_name=group_by)

        if self.column_values is not None:
            self.group_sm.layout.visibility = "visible"
            self.group_sm.layout.width = "100px"
            keep_column_values = self.column_values
            if self.column_values.dtype == np.float:
                keep_column_values = keep_column_values[~np.isnan(keep_column_values)]
            groups = np.unique(keep_column_values)
            self.group_sm.rows = min(len(groups), 20)
            self.group_sm.options = tuple(groups[::-1])
            self.group_sm.value = self.group_sm.options
            self.group_sm.disabled = False
            self.selected_column_values = self.group_sm.value
            if self.control_limit:
                self.limit_bit.disabled = False
                self.limit_cb.disabled = False
            self.group_and_sort()

    def group_dd_observer(self, change):
        """group dropdown observer"""
        if change["name"] == "value":
            group_by = change["new"]
            if group_by in ("None", "", None):
                if self.control_limit:
                    self.limit_bit.disabled = True
                    self.limit_cb.disabled = True
                    self.limit_cb.value = False
                self.column_values = None
                self.group_by = None
                self.limit = None
                self.group_sm.options = []
                self.group_sm.visible = False
                self.group_sm.rows = 1
                self.group_sm.layout.visibility = "hidden"
                self.group_sm.layout.width = "0px"

                if hasattr(self.range_controller, "slider"):
                    self.range_controller.slider.max = len(self.keep_rows)
            else:
                self.set_group_by(group_by)

            self.update_value()

    def limit_bit_observer(self, change):
        """limit bounded int text observer"""
        if change["name"] == "value":
            limit = self.limit_bit.value
            self.limit = limit
            self.update_value()

    def limit_cb_observer(self, change):
        """limit checkbox observer"""
        if not change["name"] == "value":
            return
        if self.limit_cb.value and self.group_by is not None:
            self.limit_bit.disabled = False
            self.limit = self.limit_bit.value
        else:
            self.limit_bit.disabled = True
            self.limit = None
        self.update_value()

    def order_dd_observer(self, change):
        """order dropdown observer"""
        if change["name"] == "value":
            self.order_by = self.order_dd.value

            if self.order_by is None:
                self.order_vals = range(len(self.dynamic_table))
                self.ascending_dd.disabled = True
                self.ascending_dd.layout.visibility = "hidden"
                self.ascending_dd.layout.height = "0px"
            else:
                order_vals = self.get_column_values(column_name=self.order_by)
                # convert to ints. This is mainly for handling strings
                _, order_vals = np.unique(order_vals, return_inverse=True)

                if self.desc:  # if descend is on, invert order.
                    order_vals *= -1

                self.order_vals = order_vals

                self.ascending_dd.disabled = False
                self.ascending_dd.layout.visibility = "visible"
                self.ascending_dd.layout.height = "25px"
            self.update_value()

    def ascending_dd_observer(self, change):
        """ascending dropdown observer"""
        if change["name"] == "value":
            if change["new"] == "ASC":
                self.desc = False
                self.order_vals *= -1
            else:
                self.desc = True
                self.order_vals *= -1
            self.update_value()

    def group_sm_observer(self, change):
        """group SelectMultiple observer"""
        if change["name"] == "value" and not self.group_sm.disabled:
            self.selected_column_values = change["new"]
            value_before = self.window
            self.group_and_sort()
            if (
                hasattr(self.range_controller, "slider")
                and not self.range_controller.slider.value == value_before
            ):
                pass  # do nothing, value was updated automatically
            else:
                self.update_value()

    def range_controller_observer(self, change):
        self.window = self.range_controller.value
        self.update_value()

    def get_groups(self):
        return infer_categorical_columns(self.dynamic_table, self.keep_rows)

    def get_column_values(self, column_name: str, rows_select=None):
        """Get the values of the group_by variable

        Parameters
        ----------
        column_name: str
        rows_select

        Returns
        -------

        """
        if column_name is None:
            return None
        elif column_name in self.categorical_columns:
            return (
                self.categorical_columns[column_name]
                if rows_select is None
                else self.categorical_columns[column_name][rows_select]
            )
        else:
            raise ValueError(
                "column {} not in DynamicTable {}".format(
                    column_name, self.dynamic_table
                )
            )

    def get_orderable_cols(self):
        candidate_cols = [
            x
            for x in self.units.colnames
            if not isinstance(self.units[x][0], Iterable)
            or isinstance(self.units[x][0], str)
        ]
        return [x for x in candidate_cols if len(robust_unique(self.units[x][:])) > 1]

    def group_and_sort(self):
        if self.column_values is None and self.order_vals is None:
            self.order_vals = np.arange(self.nitems).astype("int")

        order, group_inds, labels = group_and_sort(
            group_vals=self.column_values,
            group_select=self.selected_column_values,
            order_vals=self.order_vals,
            limit=self.limit,
        )

        if hasattr(self.range_controller, "slider"):
            self.range_controller.slider.max = len(order)

        # apply window
        if self.window is not None:
            order = order[self.window[0] : self.window[1]]
            if group_inds is not None:
                group_inds = group_inds[self.window[0] : self.window[1]]

        return order, group_inds, labels

    def update_value(self):

        order, group_inds, labels = self.group_and_sort()
        self.value = dict(order=order, group_inds=group_inds, labels=labels)



import numpy as np

from ipywidgets import widgets, Layout

from tqdm.notebook import tqdm as tqdm_notebook


class ProgressBar(tqdm_notebook):
    def __init__(self, *arg, **kwargs):
        super().__init__(*arg, **kwargs)
        # self.container.children[0].layout = Layout(width="80%")


def make_trial_event_controller(trials, layout=None, multiple=False):
    """Controller for which reference to use (e.g. start_time) when making time-aligned averages"""
    trial_events = ["start_time"]
    if not np.all(np.isnan(trials["stop_time"].data)):
        trial_events.append("stop_time")
    trial_events += [
        x.name
        for x in trials.columns
        if (("_time" in x.name) and (x.name not in ("start_time", "stop_time")))
    ]
    kwargs = {}
    if layout is not None:
        kwargs.update(layout=layout)

    if multiple:
        trial_event_controller = widgets.SelectMultiple(
            options=trial_events,
            value=["start_time"],
            description='align to:',
            disabled=False,
            **kwargs
        )
    else:
        trial_event_controller = widgets.Dropdown(
            options=trial_events,
            value="start_time",
            description="align to: ",
            **kwargs
        )
    return trial_event_controller



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np

import ipywidgets as widgets

from pynwb import NWBFile, TimeSeries
from pynwb.ecephys import SpikeEventSeries, ElectricalSeries, LFP

from nwbwidgets import nwb2widget
from nwbwidgets.view import default_neurodata_vis_spec
from nwbwidgets.base import show_multi_container_interface
from nwbwidgets.ecephys import show_spectrogram, show_spike_event_series, show_ccf


class ShowActivityTestCase(unittest.TestCase):
    def setUp(self):
        nwbfile = NWBFile(
            "my first synthetic recording",
            "EXAMPLE_ID",
            datetime.now(tzlocal()),
            experimenter="Dr. Matthew Douglass",
            lab="Vision Neuroscience Laboratory",
            institution="University of Middle Earth at the Shire",
            experiment_description="We recorded from macaque monkeys during memory-guided saccade task",
            session_id="LONELYMTL",
        )

        device = nwbfile.create_device(name="trodes_rig123")

        electrode_group = nwbfile.create_electrode_group(
            name="tetrode1",
            description="an example tetrode",
            location="somewhere in the hippocampus",
            device=device,
        )

        for idx in [1, 2, 3, 4]:
            nwbfile.add_electrode(
                id=idx,
                x=1.0,
                y=2.0,
                z=3.0,
                imp=float(-idx),
                location="CA1",
                filtering="none",
                group=electrode_group,
            )

        electrode_table_region = nwbfile.create_electrode_table_region(
            [0, 2], "the first and third electrodes"
        )

        self.electrodes = electrode_table_region

        rate = 10.0
        np.random.seed(1234)
        data_len = 1000
        ephys_data = np.random.rand(data_len * 2).reshape((data_len, 2))
        ephys_timestamps = np.arange(data_len) / rate

        self.ephys_ts = ElectricalSeries(
            "test_ephys_data",
            ephys_data,
            self.electrodes,
            timestamps=ephys_timestamps,
            resolution=0.001,
            description="Random numbers generated with numpy.random.rand",
        )

        self.lfp = LFP(electrical_series=self.ephys_ts, name="LFP data")

    def test_show_lfp(self):
        show_multi_container_interface(self.lfp, default_neurodata_vis_spec)

    def test_show_spike_event_series(self):
        rate = 10.0
        np.random.seed(1234)
        data_len = 1000
        ephys_data = np.random.rand(data_len * 6).reshape((data_len, 2, 3))
        ephys_timestamps = np.arange(data_len) / rate
        ses = SpikeEventSeries(
            name="test_ephys_data",
            data=ephys_data,
            timestamps=ephys_timestamps,
            electrodes=self.electrodes,
            resolution=0.001,
            comments="This data was randomly generated with numpy, using 1234 as the seed",
            description="Random numbers generated with numpy.random.rand",
        )

        assert isinstance(show_spike_event_series(ses), widgets.Widget)

    def test_show_ElectricalSeries(self):
        widget = nwb2widget(self.ephys_ts)
        widget.gas.range_controller.value = (0, 0)


def test_show_spectrogram():
    data = np.random.rand(160, 12)
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
    )

    channel = 3
    show_spectrogram(ts, channel=channel)


def test_show_ccf():
    show_ccf()



import numpy as np
import matplotlib.pyplot as plt

from nwbwidgets.utils.mpl import create_big_ax, grid_sps


def test_create_big_ax():
    fig = plt.figure(tight_layout=True)
    plt.plot(np.arange(0, 1e6, 1000))
    assert isinstance(create_big_ax(fig), plt.Subplot)


def test_grid_sps():
    shape = (5, 6)
    fig, big_ax, gs = grid_sps(shape)
    assert isinstance(fig, plt.Figure)
    assert isinstance(big_ax, plt.Subplot)
    assert isinstance(gs, plt.GridSpec)



import numbers
from typing import Iterable

import numpy as np
import pandas as pd

from pynwb.core import DynamicTable


def extract_data_from_intervals(dynamic_table: DynamicTable):
    # solution: https://stackoverflow.com/a/50297200/11483674
    # data = [x if x == x else 'NaN' for x in dynamic_table[:]]
    data = [x for x in dynamic_table[:] if x == x]
    classes = pd.unique(data).tolist()
    try:
        data = [float(i) for i in data]
        classes = [float(i) for i in classes]
        classes.sort(key=float)
    except:
        pass
    return data, classes


def infer_categorical_columns(dynamic_table: DynamicTable, region: Iterable = None):
    """
    Parameters
    ----------
    dynamic_table: DynamicTable
    region: Iterable
        row indices to select

    Returns
    -------
    categorical_cols: dict()
        keys: as columns that are categorical, values as the unique values
    """
    categorical_cols = {}
    region = (
        np.array(region, dtype="uint")
        if region is not None
        else np.arange(len(dynamic_table), dtype="uint")
    )
    for name in dynamic_table.colnames:
        if len(dynamic_table[name].shape) == 1:
            try:
                if isinstance(
                    dynamic_table[name].data[0], (str, numbers.Number, bytes)
                ):
                    column_data = [dynamic_table[name].data[i] for i in region]
                elif hasattr(dynamic_table[name].data[0], "name"):
                    column_data = [dynamic_table[name].data[i].name for i in region]
                else:
                    continue
                unique_vals = np.unique(column_data)
                if not isinstance(unique_vals[0], str) and all(np.isnan(unique_vals)):
                    continue
                if 1 < len(unique_vals) <= (len(column_data) / 2):
                    unique_vals = [
                        x.decode() if isinstance(x, bytes) else x for x in unique_vals
                    ]  # handle h5py 3.0
                    categorical_cols[name] = np.array(column_data)
            except Exception as e:
                print(e)
    return categorical_cols


def infer_columns_to_plot(dynamic_table: DynamicTable):
    """Infer which columns can be plotted in summary widgets

    Parameters
    ----------
    dynamic_table : DynamicTable

    Returns
    -------
    column_names_to_plot: list
        Columns that can be plotted with the dynamic table summary
    """

    column_names_to_plot = []

    df = dynamic_table.to_dataframe()
    categorical_columns = []
    for name in dynamic_table.colnames:
        # if name not in categorical_cols.keys():  # categorical columns can always be plotted
        value = df[name].values[0]

        if isinstance(value, (int, float, np.integer)):
            column_names_to_plot.append(name)
        elif isinstance(value, (str, bool, bytes)):
            column_names_to_plot.append(name)
            categorical_columns.append(name)

    return column_names_to_plot, categorical_columns


def group_and_sort(
    group_vals=None, group_select=None, order_vals=None, keep_rows=None, limit=None
):
    """
    Logical flow:
    0) Apply discard_rows - throw out any listed rows
    1) Apply group select - Return only values that are within this group
    2) Apply order - If group is provided, items are sorted first by group and then by order
    3) Apply limit - Applied per group
    4) Apply window - Return only items [int] through [int]. Useful for plotting

    Parameters
    ----------
    group_vals: array-like of str
        ['a', 'b', 'b', 'a', 'a', 'b', 'b', 'c']
    group_select: array-like of str
        ['a', 'b']
    order_vals: array-like of ints
        [0, 3, 4, 1, 2, 5, 6, 7]
    limit: int

    Returns
    -------

    """

    if group_vals is not None:
        if order_vals is None:
            order = np.argsort(group_vals)
        else:
            order = np.lexsort([order_vals, group_vals])
        labels, group_inds = np.unique(group_vals[order], return_inverse=True)
    else:
        labels, group_inds = None, None
        if order_vals is None:
            # cannot give order because we don't know how many units there are
            raise ValueError("group_vals and order_vals cannot both be None")
        else:
            order = np.argsort(order_vals)

    # apply discard rows
    if keep_rows is not None:
        order = order[keep_rows]
        if group_inds is not None:
            group_inds = group_inds[keep_rows]

    # apply discard NaN categories
    try:
        if any(np.isnan(labels)):
            nan_labs = np.isnan(labels)
            keep = ~np.isin(group_inds, np.where(nan_labs)[0])
            group_inds = group_inds[keep]
            order = order[keep]
            labels = labels[~np.isnan(labels)]
    except TypeError:  # if labels are strings
        pass

    if labels is not None:
        # remove groups that are missing
        labels = labels[np.isin(range(len(labels)), group_inds)]
        _, group_inds = np.unique(group_inds, return_inverse=True)

        # apply discard groups (but keep labels)
        if group_select is not None:
            keep = np.isin(labels[group_inds], group_select)
            group_inds = group_inds[keep]
            order = order[keep]

        labels = np.array([x.decode() if isinstance(x, bytes) else x for x in labels])

    # apply limit
    inds = list()
    if limit is not None:
        if group_inds is not None:
            for i in range(len(labels)):
                inds.extend(np.where(group_inds == i)[0][:limit])
            order = order[inds]
            group_inds = group_inds[inds]
        else:
            order = order[:limit]

    return order, group_inds, labels



import unittest

from datetime import datetime
from dateutil.tz import tzlocal

import numpy as np

from pynwb import NWBFile, TimeSeries
from pynwb.epoch import TimeIntervals

from nwbwidgets.utils.timeseries import (
    get_timeseries_tt,
    get_timeseries_maxt,
    get_timeseries_mint,
    get_timeseries_in_units,
    timeseries_time_to_ind,
    bisect_timeseries_by_times,
    align_by_trials,
    align_by_time_intervals,
)


def test_get_timeseries_tt():
    data = list(range(100, 200, 10))
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
    )

    tt = get_timeseries_tt(ts)
    np.testing.assert_array_equal(
        tt, [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    )


def test_get_timeseries_tt_infstarting_time():
    data = list(range(100, 200, 10))
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=np.inf, rate=1.0
    )

    tt = get_timeseries_tt(ts)
    np.testing.assert_array_equal(
        tt, [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    )


def test_get_timeseries_tt_negativeistop():
    data = list(range(100, 200, 10))
    ts = TimeSeries(
        name="test_timeseries", data=data, unit="m", starting_time=0.0, rate=1.0
    )

    tt = get_timeseries_tt(ts, istop=-1)
    np.testing.assert_array_equal(tt, [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])


def test_get_timeseries_in_units():
    data = list(range(100, 200, 10))
    timestamps = list(range(10))
    ts = TimeSeries(
        name="test_timeseries",
        data=data,
        unit="m",
        timestamps=timestamps,
        conversion=np.inf,
    )
    data, unit = get_timeseries_in_units(ts)
    assert unit is None
    assert data == [100, 110, 120, 130, 140, 150, 160, 170, 180, 190]


def test_align_by_trials():
    start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())
    create_date = datetime(2017, 4, 15, 12, tzinfo=tzlocal())

    nwbfile = NWBFile(
        session_description="NWBFile for PSTH",
        identifier="NWB123",
        session_start_time=start_time,
        file_create_date=create_date,
    )

    data = np.arange(100, 200, 10)
    timestamps = list(range(10))
    ts = TimeSeries(name="test_timeseries", data=data, unit="m", timestamps=timestamps)
    nwbfile.add_acquisition(ts)

    nwbfile.add_trial_column(
        name="stim", description="the visual stimuli during the trial"
    )
    nwbfile.add_trial(start_time=0.0, stop_time=2.0, stim="person")
    nwbfile.add_trial(start_time=3.0, stop_time=5.0, stim="ocean")
    nwbfile.add_trial(start_time=6.0, stop_time=8.0, stim="desert")

    np.testing.assert_array_equal(align_by_trials(ts), np.array([[110], [140], [170]]))


class TimeSeriesTimeStampTestCase(unittest.TestCase):
    def setUp(self):
        data = np.arange(100, 200, 10)
        timestamps = list(range(1, 4)) + list(range(7, 10)) + list(range(17, 21))
        self.ts_rate = TimeSeries(
            name="test_timeseries_rate",
            data=data,
            unit="m",
            starting_time=0.0,
            rate=1.0,
        )
        self.ts_timestamps = TimeSeries(
            name="test_timeseries_timestamps",
            data=data,
            unit="m",
            timestamps=timestamps,
        )

    def test_get_timeseries_maxt(self):
        assert get_timeseries_maxt(self.ts_rate) == 9
        assert get_timeseries_maxt(self.ts_timestamps) == 20

    def test_get_timeseries_mint(self):
        assert get_timeseries_mint(self.ts_rate) == 0
        assert get_timeseries_mint(self.ts_timestamps) == 1

    def test_timeseries_time_to_ind(self):
        assert timeseries_time_to_ind(self.ts_rate, 3.3) == 4
        assert timeseries_time_to_ind(self.ts_rate, 15.5) == 9
        assert timeseries_time_to_ind(self.ts_timestamps, 6.5) == 3
        assert timeseries_time_to_ind(self.ts_timestamps, 7.7) == 4
        assert timeseries_time_to_ind(self.ts_timestamps, 27.7) == 9

    def test_bisect_timeseries_by_times(self):
        assert np.array_equal(
            bisect_timeseries_by_times(self.ts_rate, [0, 1, 2], 4),
            [[100, 110, 120, 130], [110, 120, 130, 140], [120, 130, 140, 150]],
        )
        assert isinstance(
            bisect_timeseries_by_times(self.ts_timestamps, [0, 1, 2], 4), list
        )

    def test_get_timeseries_tt_timestamp(self):
        tt = get_timeseries_tt(self.ts_rate)
        np.testing.assert_array_equal(
            tt, [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        )

    def test_align_by_time_intervals(self):
        intervals = TimeIntervals(name="Time Intervals")
        np.testing.assert_array_equal(
            align_by_time_intervals(timeseries=self.ts_rate, intervals=intervals),
            np.array([]),
        )



"""import ipywidgets as widgets
import pynwb
from .view import default_neurodata_vis_spec
import spikeextractors as se
from pynwb.ecephys import LFP

ephys_viz_neurodata_vis_spec = dict(default_neurodata_vis_spec)


def _set_spec():
    ephys_viz_neurodata_vis_spec[pynwb.ecephys.LFP] = show_lfp


def show_lfp(node: LFP, **kwargs):
    import spikeextractors as se
    import ephys_viz as ev
    try:
        recording = LFPRecordingExtractor(lfp_node=node)
    except:
        return widgets.Text('Problem creating LFPRecordingExtractor')
    return ev.TimeseriesView(
        recording=recording,
        initial_y_scale_factor=5
    ).show(render=False)


class LFPRecordingExtractor(se.RecordingExtractor):
    def __init__(self, lfp_node: LFP):
        super().__init__()
        lfp = list(lfp_node.electrical_series.values())[0]    
        self._samplerate = lfp.rate
        self._data = lfp.data
        self._num_channels = self._data.shape[1]
        self._num_timepoints = self._data.shape[0]

    def get_channel_ids(self):
        return list(range(self._num_channels))

    def get_num_frames(self):
        return self._num_timepoints

    def get_sampling_frequency(self):
        return self._samplerate

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        return self._data[start_frame:end_frame, :][:, channel_ids].T


_set_spec()
"""
